# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
Instrumentation for AWS Lambda functions
"""

from typing import TYPE_CHECKING, Any, Callable, Dict, Tuple

if TYPE_CHECKING:
    from instana.agent.aws_lambda import AWSLambdaAgent

try:
    import sys
    import traceback

    import wrapt
    from opentelemetry.semconv.trace import SpanAttributes

    from instana import get_aws_lambda_handler
    from instana.instrumentation.aws.triggers import enrich_lambda_span, get_context
    from instana.log import logger
    from instana.singletons import env_is_aws_lambda, get_agent, get_tracer
    from instana.util.ids import define_server_timing

    def lambda_handler_with_instana(
        wrapped: Callable[..., object],
        instance: object,
        args: Tuple[object, ...],
        kwargs: Dict[str, Any],
    ) -> object:
        event = args[0]
        agent: "AWSLambdaAgent" = get_agent()
        tracer = get_tracer()

        agent.collector.collect_snapshot(*args)
        incoming_ctx = get_context(tracer, event)

        result = None
        with tracer.start_as_current_span(
            "aws.lambda.entry", span_context=incoming_ctx
        ) as span:
            enrich_lambda_span(agent, span, *args)
            try:
                result = wrapped(*args, **kwargs)

                if isinstance(result, dict):
                    server_timing_value = define_server_timing(span.context.trace_id)
                    if "headers" in result:
                        result["headers"]["Server-Timing"] = server_timing_value
                    elif "multiValueHeaders" in result:
                        result["multiValueHeaders"]["Server-Timing"] = [
                            server_timing_value
                        ]
                    if "statusCode" in result and result.get("statusCode"):
                        status_code = int(result["statusCode"])
                        span.set_attribute(SpanAttributes.HTTP_STATUS_CODE, status_code)
                        if 500 <= status_code:
                            span.record_exception(f"HTTP status {status_code}")
            except Exception as exc:
                logger.debug(f"AWS Lambda lambda_handler_with_instana error: {exc}")
                if span:
                    exc = traceback.format_exc()
                    span.record_exception(exc)
                raise
            finally:
                agent.collector.shutdown()

        if agent.collector.started:
            agent.collector.shutdown()

        return result

    if env_is_aws_lambda:
        handler_module, handler_function = get_aws_lambda_handler()

        if handler_module and handler_function:
            try:
                logger.debug(
                    f"Instrumenting AWS Lambda handler ({handler_module}.{handler_function})"
                )
                sys.path.insert(0, "/var/runtime")
                sys.path.insert(0, "/var/task")
                wrapt.wrap_function_wrapper(
                    handler_module, handler_function, lambda_handler_with_instana
                )
            except (ModuleNotFoundError, ImportError) as exc:
                logger.debug(f"AWS Lambda error: {exc}")
                logger.warning(
                    "Instana: Couldn't instrument AWS Lambda handler. Not monitoring."
                )
        else:
            logger.warning(
                "Instana: Couldn't determine AWS Lambda Handler.  Not monitoring."
            )
except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
Instrumentation for the gevent package.
"""

import sys

from opentelemetry import context
import contextvars

from instana.log import logger


def instrument_gevent():
    """Adds context propagation to gevent greenlet spawning"""
    try:
        logger.debug("Instrumenting gevent")

        import gevent

        def spawn_callback(new_greenlet):
            """Handles context propagation for newly spawning greenlets"""
            parent_context = context.get_current()
            new_context = contextvars.Context()

            new_context.run(lambda: context.attach(parent_context))
            new_greenlet.gr_context = new_context

        gevent.Greenlet.add_spawn_callback(spawn_callback)
    except Exception:
        logger.debug("instrument_gevent: ", exc_info=True)


if "gevent" not in sys.modules:
    logger.debug("Instrumenting gevent: gevent not detected or loaded.  Nothing done.")
else:
    instrument_gevent()



# (c) Copyright IBM Corp. 2025

import itertools
import os
from typing import Any, Dict, List, Sequence, Tuple, Union

from instana.configurator import config
from instana.log import logger
from instana.util.config_reader import ConfigReader

# List of supported span categories (technology or protocol)
SPAN_CATEGORIES = [
    "logging",
    "databases",
    "messaging",
    "protocols",  # http, grpc, etc.
]

# Mapping of span type calls (framework, library name, instrumentation name) to categories
SPAN_TYPE_TO_CATEGORY = {
    # Database types
    "redis": "databases",
    "mysql": "databases",
    "postgresql": "databases",
    "mongodb": "databases",
    "cassandra": "databases",
    "couchbase": "databases",
    "dynamodb": "databases",
    "sqlalchemy": "databases",
    # Messaging types
    "kafka": "messaging",
    "rabbitmq": "messaging",
    "pika": "messaging",
    "aio_pika": "messaging",
    "aioamqp": "messaging",
    # Protocol types
    "http": "protocols",
    "grpc": "protocols",
    "graphql": "protocols",
}


def parse_service_pair(pair: str) -> List[str]:
    """
    Parses a pair string to prepare a list of ignored endpoints.

    @param pair: String format:
        - "service1:method1,method2" or "service1:method1" or "service1"
    @return: List of strings in format ["service1.method1", "service1.method2", "service2.*"]
    """
    pair_list = []
    if ":" in pair:
        service, methods = pair.split(":", 1)
        service = service.strip()
        method_list = [ep.strip() for ep in methods.split(",") if ep.strip()]

        for method in method_list:
            pair_list.append(f"{service}.{method}")
    else:
        pair_list.append(f"{pair}.*")
    return pair_list


def parse_ignored_endpoints_string(params: Union[str, os.PathLike]) -> List[str]:
    """
    Parses a string to prepare a list of ignored endpoints.

    @param params: String format:
        - "service1:method1,method2;service2:method3" or "service1;service2"
    @return: List of strings in format ["service1.method1", "service1.method2", "service2.*"]
    """
    ignore_endpoints = []
    if params:
        service_pairs = params.lower().split(";")

        for pair in service_pairs:
            if pair.strip():
                ignore_endpoints += parse_service_pair(pair)
    return ignore_endpoints


def parse_ignored_endpoints_dict(params: Dict[str, Any]) -> List[str]:
    """
    Parses a dictionary to prepare a list of ignored endpoints.

    @param params: Dict format:
        - {"service1": ["method1", "method2"], "service2": ["method3"]}
    @return: List of strings in format ["service1.method1", "service1.method2", "service2.*"]
    """
    ignore_endpoints = []

    for service, methods in params.items():
        if not methods:  # filtering all service
            ignore_endpoints.append(f"{service.lower()}.*")
        else:  # filtering specific endpoints
            ignore_endpoints = parse_endpoints_of_service(
                ignore_endpoints, service, methods
            )

    return ignore_endpoints


def parse_endpoints_of_service(
    ignore_endpoints: List[str],
    service: str,
    methods: Union[str, List[str]],
) -> List[str]:
    """
    Parses endpoints of each service.

    @param ignore_endpoints: A list of rules for endpoints to be filtered.
    @param service: The name of the service to be filtered.
    @param methods: A list of specific endpoints of the service to be filtered.
    """
    if service == "kafka" and isinstance(methods, list):
        for rule in methods:
            ignore_endpoints.extend(parse_kafka_methods(rule))
    else:
        for method in methods:
            ignore_endpoints.append(f"{service.lower()}.{method.lower()}")
    return ignore_endpoints


def parse_kafka_methods(rule: Union[str, Dict[str, any]]) -> List[str]:
    parsed_rule = []
    if isinstance(rule, dict):
        for method, endpoint in itertools.product(rule["methods"], rule["endpoints"]):
            parsed_rule.append(f"kafka.{method.lower()}.{endpoint.lower()}")
    elif isinstance(rule, list):
        for method in rule:
            parsed_rule.append(f"kafka.{method.lower()}.*")
    else:
        parsed_rule.append(f"kafka.{rule.lower()}.*")
    return parsed_rule


def parse_ignored_endpoints(params: Union[Dict[str, Any], str]) -> List[str]:
    """
    Parses input to prepare a list for ignored endpoints.

    @param params: Can be either:
        - String: "service1:method1,method2;service2:method3" or "service1;service2"
        - Dict: {"service1": ["method1", "method2"], "service2": ["method3"]}
    @return: List of strings in format ["service1.method1", "service1.method2", "service2.*"]
    """
    try:
        if isinstance(params, str):
            return parse_ignored_endpoints_string(params)
        elif isinstance(params, dict):
            return parse_ignored_endpoints_dict(params)
        else:
            return []
    except Exception as e:
        logger.debug("Error parsing ignored endpoints: %s", str(e))
        return []


def parse_ignored_endpoints_from_yaml(file_path: str) -> List[str]:
    """
    Parses configuration yaml file and prepares a list of ignored endpoints.

    @param file_path: Path of the file as a string
    @return: List of strings in format ["service1.method1", "service1.method2", "service2.*", "kafka.method.topic", "kafka.*.topic", "kafka.method.*"]
    """
    config_reader = ConfigReader(file_path)
    ignore_endpoints_dict = None
    if "tracing" in config_reader.data:
        ignore_endpoints_dict = config_reader.data["tracing"].get("ignore-endpoints")
    elif "com.instana.tracing" in config_reader.data:
        logger.warning(
            'Please use "tracing" instead of "com.instana.tracing" for local configuration file.'
        )
        ignore_endpoints_dict = config_reader.data["com.instana.tracing"].get(
            "ignore-endpoints"
        )
    if ignore_endpoints_dict:
        ignored_endpoints = parse_ignored_endpoints(ignore_endpoints_dict)
        return ignored_endpoints
    else:
        return []


def is_truthy(value: Any) -> bool:
    """
    Check if a value is truthy, accepting various formats.

    @param value: The value to check
    @return: True if the value is considered truthy, False otherwise

    Accepts the following as True:
    - True (Python boolean)
    - "True", "true" (case-insensitive string)
    - "1" (string)
    - 1 (integer)
    """
    if value is None:
        return False

    if isinstance(value, bool):
        return value

    if isinstance(value, int):
        return value == 1

    if isinstance(value, str):
        value_lower = value.lower()
        return value_lower == "true" or value == "1"

    return False


def parse_span_disabling(
    disable_list: Sequence[Union[str, Dict[str, Any]]],
) -> Tuple[List[str], List[str]]:
    """
    Process a list of span disabling configurations and return lists of disabled and enabled spans.

    @param disable_list: List of span disabling configurations
    @return: Tuple of (disabled_spans, enabled_spans)
    """
    if not isinstance(disable_list, list):
        logger.debug(
            f"parse_span_disabling: Invalid disable_list type: {type(disable_list)}"
        )
        return [], []

    disabled_spans = []
    enabled_spans = []

    for item in disable_list:
        if isinstance(item, str):
            disabled = parse_span_disabling_str(item)
            disabled_spans.extend(disabled)
        elif isinstance(item, dict):
            disabled, enabled = parse_span_disabling_dict(item)
            disabled_spans.extend(disabled)
            enabled_spans.extend(enabled)
        else:
            logger.debug(
                f"parse_span_disabling: Invalid disable_list item type: {type(item)}"
            )

    return disabled_spans, enabled_spans


def parse_span_disabling_str(item: str) -> List[str]:
    """
    Process a string span disabling configuration and return a list of disabled spans.

    @param item: String span disabling configuration
    @return: List of disabled spans
    """
    if item.lower() in SPAN_CATEGORIES or item.lower() in SPAN_TYPE_TO_CATEGORY.keys():
        return [item.lower()]
    else:
        logger.debug(f"set_span_disabling_str: Invalid span category/type: {item}")
        return []


def parse_span_disabling_dict(items: Dict[str, bool]) -> Tuple[List[str], List[str]]:
    """
    Process a dictionary span disabling configuration and return lists of disabled and enabled spans.

    @param items: Dictionary span disabling configuration
    @return: Tuple of (disabled_spans, enabled_spans)
    """
    disabled_spans = []
    enabled_spans = []

    for key, value in items.items():
        if key in SPAN_CATEGORIES or key in SPAN_TYPE_TO_CATEGORY.keys():
            if is_truthy(value):
                disabled_spans.append(key)
            else:
                enabled_spans.append(key)
        else:
            logger.debug(f"set_span_disabling_dict: Invalid span category/type: {key}")

    return disabled_spans, enabled_spans


def get_disable_trace_configurations_from_env() -> Tuple[List[str], List[str]]:
    # Read INSTANA_TRACING_DISABLE environment variable
    if tracing_disable := os.environ.get("INSTANA_TRACING_DISABLE", None):
        if is_truthy(tracing_disable):
            # INSTANA_TRACING_DISABLE is True/true/1, then we disable all tracing
            disabled_spans = []
            for category in SPAN_CATEGORIES:
                disabled_spans.append(category)
            return disabled_spans, []
        else:
            # INSTANA_TRACING_DISABLE is a comma-separated list of span categories/types
            tracing_disable_list = [x.strip() for x in tracing_disable.split(",")]
            return parse_span_disabling(tracing_disable_list)
    return [], []


def get_disable_trace_configurations_from_yaml() -> Tuple[List[str], List[str]]:
    config_reader = ConfigReader(os.environ.get("INSTANA_CONFIG_PATH", ""))

    if "tracing" in config_reader.data:
        root_key = "tracing"
    elif "com.instana.tracing" in config_reader.data:
        logger.warning(
            'Please use "tracing" instead of "com.instana.tracing" for local configuration file.'
        )
        root_key = "com.instana.tracing"
    else:
        return [], []

    tracing_disable_config = config_reader.data[root_key].get("disable", "")
    return parse_span_disabling(tracing_disable_config)


def get_disable_trace_configurations_from_local() -> Tuple[List[str], List[str]]:
    if "tracing" in config:
        if tracing_disable_config := config["tracing"].get("disable", None):
            return parse_span_disabling(tracing_disable_config)
    return [], []


# Made with Bob



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2016


import os
import re
import socket
import subprocess
import sys
import threading
from typing import TYPE_CHECKING, Any, Callable

from fysom import Fysom

from instana.log import logger
from instana.util import get_default_gateway
from instana.util.process_discovery import Discovery
from instana.version import VERSION

if TYPE_CHECKING:
    from instana.agent.host import HostAgent


class TheMachine:
    RETRY_PERIOD = 30
    THREAD_NAME = "Instana Machine"

    warnedPeriodic = False

    def __init__(self, agent: "HostAgent") -> None:
        logger.debug("Initializing host agent state machine")

        self.agent = agent
        self.fsm = Fysom(
            {
                "events": [
                    ("lookup", "*", "found"),
                    ("announce", "found", "announced"),
                    ("pending", "announced", "wait4init"),
                    ("ready", "wait4init", "good2go"),
                ],
                "callbacks": {
                    # Can add the following to debug
                    # "onchangestate":  self.print_state_change,
                    "onlookup": self.lookup_agent_host,
                    "onannounce": self.announce_sensor,
                    "onpending": self.on_ready,
                    "ongood2go": self.on_good2go,
                },
            }
        )

        self.timer = threading.Timer(1, self.fsm.lookup)
        self.timer.daemon = True
        self.timer.name = self.THREAD_NAME
        self.timer.start()

    @staticmethod
    def print_state_change(e: Any) -> None:
        logger.debug(
            f"========= ({os.getpid()}#{threading.current_thread().name}) FSM event: {e.event}, src: {e.src}, dst: {e.dst} =========="
        )

    def reset(self) -> None:
        """
        reset is called to start from scratch in a process.  It may be called on first boot or
        after a detected fork.

        Here we time a new announce cycle in the future so that any existing threads have time
        to exit before we re-create them.

        :return: void
        """
        logger.debug("State machine being reset.  Will start a new announce cycle.")
        self.fsm.lookup()

    def lookup_agent_host(self, e: Any) -> bool:
        host = self.agent.options.agent_host
        port = self.agent.options.agent_port

        if self.agent.is_agent_listening(host, port):
            self.fsm.announce()
            return True

        if os.path.exists("/proc/"):
            host = get_default_gateway()
            if host:
                if self.agent.is_agent_listening(host, port):
                    self.agent.options.agent_host = host
                    self.agent.options.agent_port = port
                    self.fsm.announce()
                    return True

        if self.warnedPeriodic is False:
            logger.info(
                "Instana Host Agent couldn't be found. Will retry periodically..."
            )
            self.warnedPeriodic = True

        self.schedule_retry(
            self.lookup_agent_host, e, f"{self.THREAD_NAME}: agent_lookup"
        )
        return False

    def announce_sensor(self, e: Any) -> bool:
        logger.debug(
            f"Attempting to make an announcement to the agent on {self.agent.options.agent_host}:{self.agent.options.agent_port}"
        )
        pid = os.getpid()

        try:
            if os.path.isfile("/proc/self/cmdline"):
                with open("/proc/self/cmdline") as cmd:
                    cmdinfo = cmd.read()
                cmdline = cmdinfo.split("\x00")
            else:
                # Python doesn't provide a reliable method to determine what
                # the OS process command line may be.  Here we are forced to
                # rely on ps rather than adding a dependency on something like
                # psutil which requires dev packages, gcc etc...
                proc = subprocess.Popen(
                    ["ps", "-p", str(pid), "-o", "command"], stdout=subprocess.PIPE
                )
                (out, _) = proc.communicate()
                parts = out.split(b"\n")
                cmdline = [parts[1].decode("utf-8")]
        except Exception:
            cmdline = sys.argv
            logger.debug("announce_sensor", exc_info=True)

        d = Discovery(pid=self.__get_real_pid(), name=cmdline[0], args=cmdline[1:])

        # If we're on a system with a procfs
        if os.path.exists("/proc/"):
            try:
                # In CentOS 7, some odd things can happen such as:
                # PermissionError: [Errno 13] Permission denied: '/proc/6/fd/8'
                # Use a try/except as a safety
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.connect(
                    (self.agent.options.agent_host, self.agent.options.agent_port)
                )
                path = f"/proc/{pid}/fd/{sock.fileno()}"
                d.fd = sock.fileno()
                d.inode = os.readlink(path)
            except:  # noqa: E722
                logger.debug("Error generating file descriptor: ", exc_info=True)

        payload = self.agent.announce(d)

        if not payload:
            logger.debug("Cannot announce sensor. Scheduling retry.")
            self.schedule_retry(
                self.announce_sensor, e, f"{self.THREAD_NAME}: announce"
            )
            return False

        self.agent.set_from(payload)
        self.fsm.pending()
        logger.debug(
            f"Announced PID: {pid} (true PID: {self.agent.announce_data.pid}).  Waiting for Agent Ready..."
        )
        return True

    def schedule_retry(self, fun: Callable, e: Any, name: str) -> None:
        self.timer = threading.Timer(self.RETRY_PERIOD, fun, [e])
        self.timer.daemon = True
        self.timer.name = name
        self.timer.start()

    def on_ready(self, _: Any) -> None:
        self.agent.start()

        ns_pid = str(os.getpid())
        true_pid = str(self.agent.announce_data.pid)

        logger.info(
            f"Instana host agent available. We're in business. Announced PID: {ns_pid} (true PID: {true_pid})"
        )

    def on_good2go(self, _: Any) -> None:
        ns_pid = str(os.getpid())
        true_pid = str(self.agent.announce_data.pid)

        self.agent.log_message_to_host_agent(
            f"Instana Python Package {VERSION}: PID {ns_pid} (true PID: {true_pid}) is now online and reporting"
        )

    def __get_real_pid(self) -> int:
        """
        Attempts to determine the true process ID by querying the
        /proc/<pid>/sched file.  This works on systems with a proc filesystem.
        Otherwise default to os default.
        """
        pid = None

        if os.path.exists("/proc/"):
            sched_file = f"/proc/{os.getpid()}/sched"

            if os.path.isfile(sched_file):
                try:
                    file = open(sched_file)
                    line = file.readline()
                    g = re.search(r"\((\d+),", line)
                    if g and len(g.groups()) == 1:
                        pid = int(g.groups()[0])
                except Exception:
                    logger.debug("parsing sched file failed", exc_info=True)

        if pid is None:
            pid = os.getpid()

        return pid


# Made with Bob



# (c) Copyright IBM Corp. 2025

"""Cross-platform resource usage information"""

import os
from typing import NamedTuple

from instana.log import logger
from instana.util.runtime import is_windows


class ResourceUsage(NamedTuple):
    """
    Cross-platform resource usage information, mirroring fields found in the Unix rusage struct.

    Attributes:
        ru_utime (float): User CPU time used (seconds).
        ru_stime (float): System CPU time used (seconds).
        ru_maxrss (int): Maximum resident set size used (bytes).
        ru_ixrss (int): Integral shared memory size (bytes).
        ru_idrss (int): Integral unshared data size (bytes).
        ru_isrss (int): Integral unshared stack size (bytes).
        ru_minflt (int): Number of page reclaims (soft page faults).
        ru_majflt (int): Number of page faults requiring I/O (hard page faults).
        ru_nswap (int): Number of times a process was swapped out.
        ru_inblock (int): Number of file system input blocks.
        ru_oublock (int): Number of file system output blocks.
        ru_msgsnd (int): Number of messages sent.
        ru_msgrcv (int): Number of messages received.
        ru_nsignals (int): Number of signals received.
        ru_nvcsw (int): Number of voluntary context switches.
        ru_nivcsw (int): Number of involuntary context switches.
    """

    ru_utime: float = 0.0
    ru_stime: float = 0.0
    ru_maxrss: int = 0
    ru_ixrss: int = 0
    ru_idrss: int = 0
    ru_isrss: int = 0
    ru_minflt: int = 0
    ru_majflt: int = 0
    ru_nswap: int = 0
    ru_inblock: int = 0
    ru_oublock: int = 0
    ru_msgsnd: int = 0
    ru_msgrcv: int = 0
    ru_nsignals: int = 0
    ru_nvcsw: int = 0
    ru_nivcsw: int = 0


def get_resource_usage() -> ResourceUsage:
    """Get resource usage in a cross-platform way"""
    if is_windows():
        return _get_windows_resource_usage()
    else:
        return _get_unix_resource_usage()


def _get_unix_resource_usage() -> ResourceUsage:
    """Get resource usage on Unix systems"""
    import resource

    rusage = resource.getrusage(resource.RUSAGE_SELF)

    return ResourceUsage(
        ru_utime=rusage.ru_utime,
        ru_stime=rusage.ru_stime,
        ru_maxrss=rusage.ru_maxrss,
        ru_ixrss=rusage.ru_ixrss,
        ru_idrss=rusage.ru_idrss,
        ru_isrss=rusage.ru_isrss,
        ru_minflt=rusage.ru_minflt,
        ru_majflt=rusage.ru_majflt,
        ru_nswap=rusage.ru_nswap,
        ru_inblock=rusage.ru_inblock,
        ru_oublock=rusage.ru_oublock,
        ru_msgsnd=rusage.ru_msgsnd,
        ru_msgrcv=rusage.ru_msgrcv,
        ru_nsignals=rusage.ru_nsignals,
        ru_nvcsw=rusage.ru_nvcsw,
        ru_nivcsw=rusage.ru_nivcsw,
    )


def _get_windows_resource_usage() -> ResourceUsage:
    """Get resource usage on Windows systems"""
    # On Windows, we can use psutil to get some of the metrics
    # For metrics that aren't available, we return 0
    try:
        import psutil

        process = psutil.Process(os.getpid())

        # Get CPU times
        cpu_times = process.cpu_times()

        # Get memory info
        memory_info = process.memory_info()

        # Get IO counters
        io_counters = process.io_counters() if hasattr(process, "io_counters") else None

        # Get context switch counts if available
        ctx_switches = (
            process.num_ctx_switches() if hasattr(process, "num_ctx_switches") else None
        )

        return ResourceUsage(
            ru_utime=cpu_times.user if hasattr(cpu_times, "user") else 0.0,
            ru_stime=cpu_times.system if hasattr(cpu_times, "system") else 0.0,
            ru_maxrss=memory_info.rss // 1024
            if hasattr(memory_info, "rss")
            else 0,  # Convert to KB to match Unix
            ru_ixrss=0,  # Not available on Windows
            ru_idrss=0,  # Not available on Windows
            ru_isrss=0,  # Not available on Windows
            ru_minflt=0,  # Not directly available on Windows
            ru_majflt=0,  # Not directly available on Windows
            ru_nswap=0,  # Not available on Windows
            ru_inblock=io_counters.read_count
            if io_counters and hasattr(io_counters, "read_count")
            else 0,
            ru_oublock=io_counters.write_count
            if io_counters and hasattr(io_counters, "write_count")
            else 0,
            ru_msgsnd=0,  # Not available on Windows
            ru_msgrcv=0,  # Not available on Windows
            ru_nsignals=0,  # Not available on Windows
            ru_nvcsw=ctx_switches.voluntary
            if ctx_switches and hasattr(ctx_switches, "voluntary")
            else 0,
            ru_nivcsw=ctx_switches.involuntary
            if ctx_switches and hasattr(ctx_switches, "involuntary")
            else 0,
        )
    except ImportError:
        # If psutil is not available, return zeros
        logger.debug(
            "get_windows_resource_usage: psutil is not available, returning zeros"
        )
        return ResourceUsage()


# Made with Bob



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import os
import platform
import re
import sys
from typing import Dict, List, Tuple, Union

from instana.log import logger


def get_py_source(filename: str) -> Dict[str, str]:
    """
    Retrieves the source code for Python files requested by the UI via the host agent.

    This function reads and returns the content of Python source files. It validates
    that the requested file has a .py extension and returns an appropriate error
    message if the file cannot be read or is not a Python file.

    Args:
        filename (str): The fully qualified path to a Python source file

    Returns:
        Dict[str, str]: A dictionary containing either:
            - {"data": source_code} if successful
            - {"error": error_message} if an error occurred
    """
    response = None
    try:
        if regexp_py.search(filename) is None:
            response = {"error": "Only Python source files are allowed. (*.py)"}
        else:
            pysource = ""
            with open(filename, "r") as pyfile:
                pysource = pyfile.read()

            response = {"data": pysource}

    except Exception as exc:
        response = {"error": str(exc)}

    return response


# Used by get_py_source
regexp_py = re.compile(r"\.py$")


def determine_service_name() -> str:
    """
    Determines the most appropriate service name for this application process.

    The service name is determined using the following priority order:
    1. INSTANA_SERVICE_NAME environment variable if set
    2. For specific frameworks:
       - For gunicorn: process title or "gunicorn"
       - For Flask: FLASK_APP environment variable
       - For Django: first part of DJANGO_SETTINGS_MODULE
       - For uwsgi: "uWSGI master/worker [app_name]"
    3. Command line arguments (first non-option argument)
    4. Executable name
    5. "python" as a fallback

    Returns:
        str: The determined service name
    """
    # One environment variable to rule them all
    if "INSTANA_SERVICE_NAME" in os.environ:
        return os.environ["INSTANA_SERVICE_NAME"]

    # Now best effort in naming this process.  No nice package.json like in Node.js
    # so we do best effort detection here.
    app_name = "python"  # the default name
    basename = None

    try:
        if not hasattr(sys, "argv"):
            proc_cmdline = get_proc_cmdline(as_string=False)
            return os.path.basename(proc_cmdline[0])

        # Get first argument that is not an CLI option
        for candidate in sys.argv:
            if len(candidate) > 0 and candidate[0] != "-":
                basename = candidate
                break

        # If nothing found, fall back to executable
        if basename is None:
            basename = os.path.basename(sys.executable)
        else:
            # Assure leading paths are stripped
            basename = os.path.basename(basename)

        if basename == "gunicorn":
            if "setproctitle" in sys.modules:
                # With the setproctitle package, gunicorn renames their processes
                # to pretty things - we use those by default
                # gunicorn: master [djface.wsgi]
                # gunicorn: worker [djface.wsgi]
                app_name = get_proc_cmdline(as_string=True)
            else:
                app_name = basename
        elif "FLASK_APP" in os.environ:
            app_name = os.environ["FLASK_APP"]
        elif "DJANGO_SETTINGS_MODULE" in os.environ:
            app_name = os.environ["DJANGO_SETTINGS_MODULE"].split(".")[0]
        elif basename == "":
            if sys.stdout.isatty():
                app_name = "Interactive Console"
            else:
                # No arguments.  Take executable as app_name
                app_name = os.path.basename(sys.executable)
        else:
            # Last chance.  app_name for "python main.py" would be "main.py" here.
            app_name = basename

        # We should have a good app_name by this point.
        # Last conditional, if uwsgi, then wrap the name
        # with the uwsgi process type
        if basename == "uwsgi":
            # We have an app name by this point.  Now if running under
            # uwsgi, augment the app name
            try:
                import uwsgi

                if app_name == "uwsgi":
                    app_name = ""
                else:
                    app_name = " [%s]" % app_name

                if os.getpid() == uwsgi.masterpid():
                    uwsgi_type = "uWSGI master%s"
                else:
                    uwsgi_type = "uWSGI worker%s"

                app_name = uwsgi_type % app_name
            except (ImportError, AttributeError):
                pass
    except Exception:
        logger.debug("non-fatal get_application_name: ", exc_info=True)

    return app_name


def get_proc_cmdline(as_string: bool = False) -> Union[List[str], str]:
    """
    Parses the process command line from the proc file system.

    This function attempts to read the command line of the current process from
    /proc/self/cmdline. If the proc filesystem is not available (e.g., on non-Unix
    systems), it returns a default value.

    Args:
        as_string (bool, optional): If True, returns the command line as a single
                                   space-separated string. If False, returns a list
                                   of command line arguments. Defaults to False.

    Returns:
        Union[List[str], str]: The command line as either a list of arguments or a
                              space-separated string, depending on the as_string parameter.
    """
    name = "python"
    if os.path.isfile("/proc/self/cmdline"):
        with open("/proc/self/cmdline") as cmd:
            name = cmd.read()
    else:
        # Most likely not on a *nix based OS.  Return a default
        if as_string is True:
            return name
        else:
            return [name]

    # /proc/self/command line will have strings with null bytes such as "/usr/bin/python\0-s\0-d\0".  This
    # bit will prep the return value and drop the trailing null byte
    parts = name.split("\0")
    parts.pop()

    if as_string is True:
        parts = " ".join(parts)

    return parts


def get_runtime_env_info() -> Tuple[str, str, str]:
    """
    Returns information about the current runtime environment.

    This function collects and returns details about the machine architecture
    and Python version being used by the application.

    Returns:
        Tuple[str, str, str]: A tuple containing:
            - Machine type (e.g., 'arm64', 'ppc64le')
            - System/OS name (e.g., ' Linux', 'Windows')
            - Python version string
    """
    machine = platform.machine()
    system = platform.system()
    python_version = platform.python_version()

    return machine, system, python_version


def log_runtime_env_info() -> None:
    """
    Logs debug information about the current runtime environment.

    This function retrieves machine architecture and Python version information
    using get_runtime_env_info() and logs it as a debug message.
    """
    machine, system, python_version = get_runtime_env_info()
    logger.debug(
        f"Runtime environment: Machine: {machine}, System: {system}, Python version: {python_version}"
    )


def is_windows() -> bool:
    """
    Checks if the current runtime environment is running on a Windows operating system.

    Returns:
        bool: True if the current runtime environment is Windows, False otherwise.
    """
    system = get_runtime_env_info()[1].lower()
    return system == "windows"


def is_ppc64() -> bool:
    """
    Checks if the current runtime environment is running on ppc64 architecture.

    Returns:
        bool: True if the current runtime environment is on ppc64 architecture, False otherwise.
    """
    machine = get_runtime_env_info()[0].lower()
    return machine.startswith("ppc64")


def is_s390x() -> bool:
    """
    Checks if the current runtime environment is running on s390x architecture.

    Returns:
        bool: True if the current runtime environment is on s390x architecture, False otherwise.
    """
    machine = get_runtime_env_info()[0].lower()
    return machine == "s390x"


# Made with Bob



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import os
import pytest

import urllib3
import gevent
from gevent.pool import Group
from typing import Generator

import tests.apps.flask_app
from instana.singletons import tracer
from tests.helpers import testenv, get_spans_by_filter, filter_test_span


# Skip the tests if the environment variable `GEVENT_TEST` is not set
pytestmark = pytest.mark.skipif(not os.environ.get("GEVENT_TEST"), reason="GEVENT_TEST not set")


class TestGEvent:
    @classmethod
    def setup_class(cls) -> None:
        """Setup that runs once before all tests in the class"""
        cls.http = urllib3.HTTPConnectionPool('127.0.0.1', port=testenv["flask_port"], maxsize=20)
        cls.recorder = tracer.span_processor

    @pytest.fixture(autouse=True)
    def setUp(self) -> Generator[None, None, None]:
        """Clear all spans before each test run"""
        self.recorder.clear_spans()

    def make_http_call(self, n=None):
        """Helper function to make HTTP calls"""
        return self.http.request('GET', testenv["flask_server"] + '/')

    def spawn_calls(self):
        """Helper function to spawn multiple HTTP calls"""
        with tracer.start_as_current_span('spawn_calls'):
            jobs = []
            jobs.append(gevent.spawn(self.make_http_call))
            jobs.append(gevent.spawn(self.make_http_call))
            jobs.append(gevent.spawn(self.make_http_call))
            gevent.joinall(jobs, timeout=2)

    def spawn_imap_unordered(self):
        """Helper function to test imap_unordered"""
        igroup = Group()
        result = []
        with tracer.start_as_current_span('test'):
            for i in igroup.imap_unordered(self.make_http_call, range(3)):
                result.append(i)

    def launch_gevent_chain(self):
        """Helper function to launch a chain of gevent calls"""
        with tracer.start_as_current_span('test'):
            gevent.spawn(self.spawn_calls).join()

    def test_spawning(self):
        gevent.spawn(self.launch_gevent_chain)
        gevent.sleep(2)
        
        spans = self.recorder.queued_spans()
        
        assert len(spans) == 8
        
        test_spans = get_spans_by_filter(spans, filter_test_span)
        assert test_spans
        assert len(test_spans) == 1
        
        test_span = test_spans[0]
        
        span_filter = lambda span: span.n == "sdk" \
                                  and span.data['sdk']['name'] == 'spawn_calls' and span.p == test_span.s
        spawn_spans = get_spans_by_filter(spans, span_filter)
        assert spawn_spans
        assert len(spawn_spans) == 1
        
        spawn_span = spawn_spans[0]
        
        span_filter = lambda span: span.n == "urllib3"
        urllib3_spans = get_spans_by_filter(spans, span_filter)
        
        for urllib3_span in urllib3_spans:
            # spans should all have the same test span parent
            assert urllib3_span.t == spawn_span.t
            assert urllib3_span.p == spawn_span.s
            
            # find the wsgi span generated from this urllib3 request
            span_filter = lambda span: span.n == "wsgi" and span.p == urllib3_span.s
            wsgi_spans = get_spans_by_filter(spans, span_filter)
            assert wsgi_spans is not None
            assert len(wsgi_spans) == 1

    def test_imap_unordered(self):
        gevent.spawn(self.spawn_imap_unordered)
        gevent.sleep(2)
        
        spans = self.recorder.queued_spans()
        assert len(spans) == 7
        
        test_spans = get_spans_by_filter(spans, filter_test_span)
        assert test_spans is not None
        assert len(test_spans) == 1
        
        test_span = test_spans[0]
        
        span_filter = lambda span: span.n == "urllib3"
        urllib3_spans = get_spans_by_filter(spans, span_filter)
        assert len(urllib3_spans) == 3
        
        for urllib3_span in urllib3_spans:
            # spans should all have the same test span parent
            assert urllib3_span.t == test_span.t
            assert urllib3_span.p == test_span.s
            
            # find the wsgi span generated from this urllib3 request
            span_filter = lambda span: span.n == "wsgi" and span.p == urllib3_span.s
            wsgi_spans = get_spans_by_filter(spans, span_filter)
            assert wsgi_spans is not None
            assert len(wsgi_spans) == 1



# (c) Copyright IBM Corp. 2025

try:
    import httpx
    import wrapt
    from typing import TYPE_CHECKING, Any, Callable, Dict, Tuple, Optional
    from opentelemetry.semconv.trace import SpanAttributes
    from opentelemetry.trace import SpanKind

    from instana.log import logger
    from instana.propagators.format import Format
    from instana.singletons import agent
    from instana.util.secrets import strip_secrets_from_query
    from instana.util.traceutils import (
        extract_custom_headers,
        get_tracer_tuple,
        tracing_is_off,
    )

    if TYPE_CHECKING:
        from instana.span.span import InstanaSpan

    def _set_request_span_attributes(
        span: "InstanaSpan",
        request: httpx.Request,
    ) -> None:
        try:
            url = request.url

            # Strip any secrets from potential query params
            if url.query:
                formatted_query = strip_secrets_from_query(
                    str(url.query, encoding="utf-8"),
                    agent.options.secrets_matcher,
                    agent.options.secrets_list,
                )
                span.set_attribute("http.params", formatted_query)

            url_str = f"{url.scheme}://{url.host}"
            if url.port:
                url_str += f":{url.port}"
            url_str += f"{url.path}"

            span.set_attribute(SpanAttributes.HTTP_URL, url_str)
            span.set_attribute(SpanAttributes.HTTP_HOST, url.host)
            span.set_attribute(SpanAttributes.HTTP_METHOD, request.method)
            span.set_attribute("http.path", url.path)

            extract_custom_headers(span, request.headers)
        except Exception:
            logger.debug("httpx _set_request_span_attributes error: ", exc_info=True)

    def _set_response_span_attributes(
        span: "InstanaSpan",
        response: Optional[httpx.Response] = None,
    ) -> None:
        try:
            if response.headers:
                extract_custom_headers(span, response.headers)

            status_code = response.status_code
            span.set_attribute(SpanAttributes.HTTP_STATUS_CODE, status_code)
            if 500 <= status_code:
                span.mark_as_errored()
        except Exception:
            logger.debug("httpx _set_request_span_attributes error: ", exc_info=True)

    @wrapt.patch_function_wrapper("httpx", "HTTPTransport.handle_request")
    def handle_request_with_instana(
        wrapped: Callable[..., "httpx.HTTPTransport.handle_request"],
        instance: httpx.HTTPTransport,
        args: Tuple[int, str, Tuple[Any, ...]],
        kwargs: Dict[str, Any],
    ) -> httpx.Response:
        # If we're not tracing, just return
        if tracing_is_off():
            return wrapped(*args, **kwargs)

        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span(
            "httpx", span_context=parent_context, kind=SpanKind.CLIENT
        ) as span:
            try:
                request = args[0]
                _set_request_span_attributes(span, request)
                tracer.inject(span.context, Format.HTTP_HEADERS, request.headers)

                response = wrapped(*args, **kwargs)
                _set_response_span_attributes(span, response)
            except Exception as e:
                span.record_exception(e)
            else:
                return response

    @wrapt.patch_function_wrapper("httpx", "AsyncHTTPTransport.handle_async_request")
    async def handle_async_request_with_instana(
        wrapped: Callable[..., "httpx.AsyncHTTPTransport.handle_async_request"],
        instance: httpx.AsyncHTTPTransport,
        args: Tuple[int, str, Tuple[Any, ...]],
        kwargs: Dict[str, Any],
    ) -> httpx.Response:
        # If we're not tracing, just return
        if tracing_is_off():
            return await wrapped(*args, **kwargs)

        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span(
            "httpx", span_context=parent_context, kind=SpanKind.CLIENT
        ) as span:
            try:
                request = args[0]
                _set_request_span_attributes(span, request)
                tracer.inject(span.context, Format.HTTP_HEADERS, request.headers)

                response = await wrapped(*args, **kwargs)
                _set_response_span_attributes(span, response)
            except Exception as e:
                span.record_exception(e)
            else:
                return response

    logger.debug("Instrumenting httpx")

except ImportError:
    pass



# (c) Copyright IBM Corp. 2025

from typing import TYPE_CHECKING, Any, Callable, Dict, Sequence, Type

if TYPE_CHECKING:
    from botocore.client import BaseClient

from instana.log import logger
from instana.singletons import tracer
from instana.span_context import SpanContext


def create_dynamodb_span(
    wrapped: Callable[..., Dict[str, Any]],
    instance: Type["BaseClient"],
    args: Sequence[Dict[str, Any]],
    kwargs: Dict[str, Any],
    parent_context: SpanContext,
) -> None:
    with tracer.start_as_current_span("dynamodb", span_context=parent_context) as span:
        try:
            span.set_attribute("dynamodb.op", args[0])
            span.set_attribute("dynamodb.region", instance._client_config.region_name)
            if "TableName" in args[1].keys():
                span.set_attribute("dynamodb.table", args[1]["TableName"])
        except Exception as exc:
            span.record_exception(exc)
            logger.debug("create_dynamodb_span: collect error", exc_info=True)


logger.debug("Instrumenting DynamoDB")



# (c) Copyright IBM Corp. 2025

import pytest

from instana.util.config import (is_truthy, parse_endpoints_of_service,
                                 parse_ignored_endpoints,
                                 parse_ignored_endpoints_dict,
                                 parse_kafka_methods, parse_service_pair)


class TestConfig:
    def test_parse_service_pair(self) -> None:
        test_string = "service1:method1,method2"
        response = parse_service_pair(test_string)
        assert response == ["service1.method1", "service1.method2"]

        test_string = "service1;service2"
        response = parse_ignored_endpoints(test_string)
        assert response == ["service1.*", "service2.*"]

        test_string = "service1"
        response = parse_ignored_endpoints(test_string)
        assert response == ["service1.*"]

        test_string = ";"
        response = parse_ignored_endpoints(test_string)
        assert response == []

        test_string = "service1:method1,method2;;;service2:method1;;"
        response = parse_ignored_endpoints(test_string)
        assert response == [
            "service1.method1",
            "service1.method2",
            "service2.method1",
        ]

        test_string = ""
        response = parse_ignored_endpoints(test_string)
        assert response == []

    def test_parse_ignored_endpoints_string(self) -> None:
        test_string = "service1:method1,method2"
        response = parse_service_pair(test_string)
        assert response == ["service1.method1", "service1.method2"]

        test_string = "service1;service2"
        response = parse_ignored_endpoints(test_string)
        assert response == ["service1.*", "service2.*"]

        test_string = "service1"
        response = parse_ignored_endpoints(test_string)
        assert response == ["service1.*"]

        test_string = ";"
        response = parse_ignored_endpoints(test_string)
        assert response == []

        test_string = "service1:method1,method2;;;service2:method1;;"
        response = parse_ignored_endpoints(test_string)
        assert response == [
            "service1.method1",
            "service1.method2",
            "service2.method1",
        ]

        test_string = ""
        response = parse_ignored_endpoints(test_string)
        assert response == []

    def test_parse_ignored_endpoints_dict(self) -> None:
        test_dict = {"service1": ["method1", "method2"]}
        response = parse_ignored_endpoints_dict(test_dict)
        assert response == ["service1.method1", "service1.method2"]

        test_dict = {"SERVICE1": ["method1", "method2"]}
        response = parse_ignored_endpoints_dict(test_dict)
        assert response == ["service1.method1", "service1.method2"]

        test_dict = {"service1": [], "service2": []}
        response = parse_ignored_endpoints_dict(test_dict)
        assert response == ["service1.*", "service2.*"]

        test_dict = {"service1": []}
        response = parse_ignored_endpoints_dict(test_dict)
        assert response == ["service1.*"]

        test_dict = {}
        response = parse_ignored_endpoints_dict(test_dict)
        assert response == []

    def test_parse_ignored_endpoints(self) -> None:
        test_pair = "service1:method1,method2"
        response = parse_ignored_endpoints(test_pair)
        assert response == ["service1.method1", "service1.method2"]

        test_pair = "service1;service2"
        response = parse_ignored_endpoints(test_pair)
        assert response == ["service1.*", "service2.*"]

        test_pair = "service1"
        response = parse_ignored_endpoints(test_pair)
        assert response == ["service1.*"]

        test_pair = ";"
        response = parse_ignored_endpoints(test_pair)
        assert response == []

        test_pair = "service1:method1,method2;;;service2:method1;;"
        response = parse_ignored_endpoints(test_pair)
        assert response == [
            "service1.method1",
            "service1.method2",
            "service2.method1",
        ]

        test_pair = ""
        response = parse_ignored_endpoints(test_pair)
        assert response == []

        test_dict = {"service1": ["method1", "method2"]}
        response = parse_ignored_endpoints(test_dict)
        assert response == ["service1.method1", "service1.method2"]

        test_dict = {"service1": [], "service2": []}
        response = parse_ignored_endpoints(test_dict)
        assert response == ["service1.*", "service2.*"]

        test_dict = {"service1": []}
        response = parse_ignored_endpoints(test_dict)
        assert response == ["service1.*"]

        test_dict = {}
        response = parse_ignored_endpoints(test_dict)
        assert response == []

    def test_parse_endpoints_of_service(self) -> None:
        test_ignore_endpoints = {
            "service1": ["method1", "method2"],
            "service2": ["method3", "method4"],
            "kafka": [
                {
                    "methods": ["method5", "method6"],
                    "endpoints": ["endpoint1", "endpoint2"],
                }
            ],
        }
        ignore_endpoints = []
        for service, methods in test_ignore_endpoints.items():
            ignore_endpoints.extend(parse_endpoints_of_service([], service, methods))
        assert ignore_endpoints == [
            "service1.method1",
            "service1.method2",
            "service2.method3",
            "service2.method4",
            "kafka.method5.endpoint1",
            "kafka.method5.endpoint2",
            "kafka.method6.endpoint1",
            "kafka.method6.endpoint2",
        ]

    def test_parse_kafka_methods_as_dict(self) -> None:
        test_rule_as_dict = {"methods": ["send"], "endpoints": ["topic1"]}
        parsed_rule = parse_kafka_methods(test_rule_as_dict)
        assert parsed_rule == ["kafka.send.topic1"]

    def test_parse_kafka_methods_as_str(self) -> None:
        test_rule_as_str = ["send"]
        parsed_rule = parse_kafka_methods(test_rule_as_str)
        assert parsed_rule == ["kafka.send.*"]
        
    @pytest.mark.parametrize("value, expected", [
        (True, True),
        (False, False),
        ("True", True),
        ("true", True),
        ("1", True),
        (1, True),
        ("False", False),
        ("false", False),
        ("0", False),
        (0, False),
        (None, False),
        ("TRUE", True),
        ("FALSE", False),
        ("yes", False),  # Only "true" and "1" are considered truthy
        ("no", False),
    ])
    def test_is_truthy(self, value, expected) -> None:
        """Test the is_truthy function with various input values."""
        assert is_truthy(value) == expected



# (c) Copyright IBM Corp. 2025

try:
    import aioamqp
    from typing import Any, Callable, Dict, Tuple

    import wrapt
    from opentelemetry.trace.status import StatusCode

    from instana.log import logger
    from instana.util.traceutils import get_tracer_tuple, tracing_is_off

    @wrapt.patch_function_wrapper("aioamqp.channel", "Channel.basic_publish")
    async def basic_publish_with_instana(
        wrapped: Callable[..., aioamqp.connect],
        instance: object,
        argv: Tuple[object, Tuple[object, ...]],
        kwargs: Dict[str, Any],
    ) -> object:
        if tracing_is_off():
            return await wrapped(*argv, **kwargs)

        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None
        with tracer.start_as_current_span(
            "aioamqp-publisher", span_context=parent_context
        ) as span:
            try:
                span.set_attribute("amqp.command", "publish")
                span.set_attribute("amqp.routing_key", kwargs.get("routing_key"))

                protocol = getattr(instance, "protocol", None)
                transport = getattr(protocol, "_transport", None)
                extra = getattr(transport, "_extra", {}) if transport else {}
                peername = extra.get("peername")
                if (
                    peername
                    and isinstance(peername, (list, tuple))
                    and len(peername) >= 2
                ):
                    connection_info = f"{peername[0]}:{peername[1]}"
                else:
                    connection_info = "unknown"
                span.set_attribute("amqp.connection", connection_info)

                response = await wrapped(*argv, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
                logger.debug(f"aioamqp basic_publish_with_instana error: {exc}")
            else:
                return response

    @wrapt.patch_function_wrapper("aioamqp.channel", "Channel.basic_consume")
    async def basic_consume_with_instana(
        wrapped: Callable[..., aioamqp.connect],
        instance: object,
        argv: Tuple[object, Tuple[object, ...]],
        kwargs: Dict[str, Any],
    ) -> object:
        if tracing_is_off():
            return await wrapped(*argv, **kwargs)

        callback = argv[0]
        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None

        @wrapt.decorator
        async def callback_wrapper(
            wrapped_callback: Callable[..., aioamqp.connect],
            instance: Any,
            args: Tuple,
            kwargs: Dict,
        ) -> object:
            with tracer.start_as_current_span(
                "aioamqp-consumer", span_context=parent_context
            ) as span:
                try:
                    span.set_status(StatusCode.OK)
                    span.set_attribute("amqp.command", "consume")
                    span.set_attribute("amqp.routing_key", args[2].routing_key)

                    protocol = getattr(args[0], "protocol", None)
                    transport = getattr(protocol, "_transport", None)
                    extra = getattr(transport, "_extra", {}) if transport else {}
                    peername = extra.get("peername")
                    if (
                        peername
                        and isinstance(peername, (list, tuple))
                        and len(peername) >= 2
                    ):
                        connection_info = f"{peername[0]}:{peername[1]}"
                    else:
                        connection_info = "unknown"
                    span.set_attribute("amqp.connection", connection_info)

                    response = await wrapped_callback(*args, **kwargs)
                except Exception as exc:
                    span.record_exception(exc)
                    logger.debug(f"aioamqp basic_consume_with_instana error: {exc}")
                else:
                    return response

        wrapped_callback = callback_wrapper(callback)
        argv = (wrapped_callback,) + argv[1:]

        return await wrapped(*argv, **kwargs)

    logger.debug("Instrumenting aioamqp")

except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020


import logging
import os
from typing import Generator
from unittest.mock import patch

import pytest
import redis

from instana.options import StandardOptions
from instana.singletons import agent, tracer
from instana.span.span import get_current_span
from tests.helpers import testenv


class TestRedis:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """Clear all spans before a test run"""
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        self.client = redis.Redis(host=testenv["redis_host"], db=testenv["redis_db"])
        yield
        if "INSTANA_IGNORE_ENDPOINTS" in os.environ.keys():
            del os.environ["INSTANA_IGNORE_ENDPOINTS"]
        agent.options.allow_exit_as_root = False

    def test_set_get(self) -> None:
        result = None
        with tracer.start_as_current_span("test"):
            self.client.set("foox", "barX")
            self.client.set("fooy", "barY")
            result = self.client.get("foox")

        spans = self.recorder.queued_spans()
        assert len(spans) == 4

        assert result == b"barX"

        rs1_span = spans[0]
        rs2_span = spans[1]
        rs3_span = spans[2]
        test_span = spans[3]

        current_span = get_current_span()
        assert not current_span.is_recording()

        # Same traceId
        assert rs1_span.t == test_span.t
        assert rs2_span.t == test_span.t
        assert rs3_span.t == test_span.t

        # Parent relationships
        assert rs1_span.p == test_span.s
        assert rs2_span.p == test_span.s
        assert rs3_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not rs1_span.ec
        assert not rs2_span.ec
        assert not rs3_span.ec

        # Redis span 1
        assert rs1_span.n == "redis"
        assert "custom" not in rs1_span.data
        assert "redis" in rs1_span.data

        assert rs1_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs1_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs1_span.data["redis"]["command"] == "SET"
        assert not rs1_span.data["redis"]["error"]

        assert rs1_span.stack
        assert isinstance(rs1_span.stack, list)
        assert len(rs1_span.stack) > 0

        # Redis span 2
        assert rs2_span.n == "redis"
        assert "custom" not in rs2_span.data
        assert "redis" in rs2_span.data

        assert rs2_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs2_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs2_span.data["redis"]["command"] == "SET"
        assert not rs2_span.data["redis"]["error"]

        assert rs2_span.stack
        assert isinstance(rs2_span.stack, list)
        assert len(rs2_span.stack) > 0

        # Redis span 3
        assert rs3_span.n == "redis"
        assert "custom" not in rs3_span.data
        assert "redis" in rs3_span.data

        assert rs3_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs3_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs3_span.data["redis"]["command"] == "GET"
        assert not rs3_span.data["redis"]["error"]

        assert rs3_span.stack
        assert isinstance(rs3_span.stack, list)
        assert len(rs3_span.stack) > 0

    def test_set_get_as_root_span(self) -> None:
        agent.options.allow_exit_as_root = True

        self.client.set("foox", "barX")
        self.client.set("fooy", "barY")
        result = self.client.get("foox")

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        assert result == b"barX"

        rs1_span = spans[0]
        rs2_span = spans[1]
        rs3_span = spans[2]

        current_span = get_current_span()
        assert not current_span.is_recording()

        # Parent relationships
        assert not rs1_span.p
        assert not rs2_span.p
        assert not rs3_span.p

        # Error logging
        assert not rs1_span.ec
        assert not rs2_span.ec
        assert not rs3_span.ec

        # Redis span 1
        assert rs1_span.n == "redis"
        assert "custom" not in rs1_span.data
        assert "redis" in rs1_span.data

        assert rs1_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs1_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs1_span.data["redis"]["command"] == "SET"
        assert not rs1_span.data["redis"]["error"]

        assert rs1_span.stack
        assert isinstance(rs1_span.stack, list)
        assert len(rs1_span.stack) > 0

        # Redis span 2
        assert rs2_span.n == "redis"
        assert "custom" not in rs2_span.data
        assert "redis" in rs2_span.data

        assert rs2_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs2_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs2_span.data["redis"]["command"] == "SET"
        assert not rs2_span.data["redis"]["error"]

        assert rs2_span.stack
        assert isinstance(rs2_span.stack, list)
        assert len(rs2_span.stack) > 0

        # Redis span 3
        assert rs3_span.n == "redis"
        assert "custom" not in rs3_span.data
        assert "redis" in rs3_span.data

        assert rs3_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs3_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs3_span.data["redis"]["command"] == "GET"
        assert not rs3_span.data["redis"]["error"]

        assert rs3_span.stack
        assert isinstance(rs3_span.stack, list)
        assert len(rs3_span.stack) > 0

    def test_set_incr_get(self) -> None:
        result = None
        with tracer.start_as_current_span("test"):
            self.client.set("counter", "10")
            self.client.incr("counter")
            result = self.client.get("counter")

        spans = self.recorder.queued_spans()
        assert len(spans) == 4

        assert result == b"11"

        rs1_span = spans[0]
        rs2_span = spans[1]
        rs3_span = spans[2]
        test_span = spans[3]

        current_span = get_current_span()
        assert not current_span.is_recording()

        # Same traceId
        assert rs1_span.t == test_span.t
        assert rs2_span.t == test_span.t
        assert rs3_span.t == test_span.t

        # Parent relationships
        assert rs1_span.p == test_span.s
        assert rs2_span.p == test_span.s
        assert rs3_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not rs1_span.ec
        assert not rs2_span.ec
        assert not rs3_span.ec

        # Redis span 1
        assert rs1_span.n == "redis"
        assert "custom" not in rs1_span.data
        assert "redis" in rs1_span.data

        assert rs1_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs1_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs1_span.data["redis"]["command"] == "SET"
        assert not rs1_span.data["redis"]["error"]

        assert rs1_span.stack
        assert isinstance(rs1_span.stack, list)
        assert len(rs1_span.stack) > 0

        # Redis span 2
        assert rs2_span.n == "redis"
        assert "custom" not in rs2_span.data
        assert "redis" in rs2_span.data

        assert rs2_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs2_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs2_span.data["redis"]["command"] == "INCRBY"
        assert not rs2_span.data["redis"]["error"]

        assert rs2_span.stack
        assert isinstance(rs2_span.stack, list)
        assert len(rs2_span.stack) > 0

        # Redis span 3
        assert rs3_span.n == "redis"
        assert "custom" not in rs3_span.data
        assert "redis" in rs3_span.data

        assert rs3_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs3_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs3_span.data["redis"]["command"] == "GET"
        assert not rs3_span.data["redis"]["error"]

        assert rs3_span.stack
        assert isinstance(rs3_span.stack, list)
        assert len(rs3_span.stack) > 0

    def test_old_redis_client(self) -> None:
        result = None
        with tracer.start_as_current_span("test"):
            self.client.set("foox", "barX")
            self.client.set("fooy", "barY")
            result = self.client.get("foox")

        spans = self.recorder.queued_spans()
        assert len(spans) == 4

        assert result == b"barX"

        rs1_span = spans[0]
        rs2_span = spans[1]
        rs3_span = spans[2]
        test_span = spans[3]

        current_span = get_current_span()
        assert not current_span.is_recording()

        # Same traceId
        assert rs1_span.t == test_span.t
        assert rs2_span.t == test_span.t
        assert rs3_span.t == test_span.t

        # Parent relationships
        assert rs1_span.p == test_span.s
        assert rs2_span.p == test_span.s
        assert rs3_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not rs1_span.ec
        assert not rs2_span.ec
        assert not rs3_span.ec

        # Redis span 1
        assert rs1_span.n == "redis"
        assert "custom" not in rs1_span.data
        assert "redis" in rs1_span.data

        assert rs1_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs1_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs1_span.data["redis"]["command"] == "SET"
        assert not rs1_span.data["redis"]["error"]

        assert rs1_span.stack
        assert isinstance(rs1_span.stack, list)
        assert len(rs1_span.stack) > 0

        # Redis span 2
        assert rs2_span.n == "redis"
        assert "custom" not in rs2_span.data
        assert "redis" in rs2_span.data

        assert rs2_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs2_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )

        assert rs2_span.data["redis"]["command"] == "SET"
        assert not rs2_span.data["redis"]["error"]

        assert rs2_span.stack
        assert isinstance(rs2_span.stack, list)
        assert len(rs2_span.stack) > 0

        # Redis span 3
        assert rs3_span.n == "redis"
        assert "custom" not in rs3_span.data
        assert "redis" in rs3_span.data

        assert rs3_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs3_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs3_span.data["redis"]["command"] == "GET"
        assert not rs3_span.data["redis"]["error"]

        assert rs3_span.stack
        assert isinstance(rs3_span.stack, list)
        assert len(rs3_span.stack) > 0

    def test_pipelined_requests(self) -> None:
        result = None
        with tracer.start_as_current_span("test"):
            pipe = self.client.pipeline()
            pipe.set("foox", "barX")
            pipe.set("fooy", "barY")
            pipe.get("foox")
            result = pipe.execute()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        assert result == [True, True, b"barX"]

        rs1_span = spans[0]
        test_span = spans[1]

        current_span = get_current_span()
        assert not current_span.is_recording()

        # Same traceId
        assert rs1_span.t == test_span.t

        # Parent relationships
        assert rs1_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not rs1_span.ec

        # Redis span 1
        assert rs1_span.n == "redis"
        assert "custom" not in rs1_span.data
        assert "redis" in rs1_span.data

        assert rs1_span.data["redis"]["driver"] == "redis-py"
        assert (
            rs1_span.data["redis"]["connection"]
            == f"redis://{testenv['redis_host']}:6379/0"
        )
        assert rs1_span.data["redis"]["command"] == "PIPELINE"
        assert rs1_span.data["redis"]["subCommands"] == ["SET", "SET", "GET"]
        assert not rs1_span.data["redis"]["error"]

        assert rs1_span.stack
        assert isinstance(rs1_span.stack, list)
        assert len(rs1_span.stack) > 0

    @patch(
        "instana.instrumentation.redis.collect_attributes",
        side_effect=Exception("test-error"),
    )
    @patch("instana.span.span.InstanaSpan.record_exception")
    def test_execute_command_with_instana_exception(self, mock_record_func, _) -> None:
        with tracer.start_as_current_span("test"), pytest.raises(
            Exception, match="test-error"
        ):
            self.client.set("counter", "10")
        mock_record_func.assert_called()

    def test_execute_comand_with_instana_tracing_off(self) -> None:
        with tracer.start_as_current_span("redis"):
            response = self.client.set("counter", "10")
            assert response

    def test_execute_with_instana_tracing_off(self) -> None:
        result = None
        with tracer.start_as_current_span("redis"):
            pipe = self.client.pipeline()
            pipe.set("foox", "barX")
            pipe.set("fooy", "barY")
            pipe.get("foox")
            result = pipe.execute()
            assert result == [True, True, b"barX"]

    def test_execute_with_instana_exception(
        self, caplog: pytest.LogCaptureFixture
    ) -> None:
        caplog.set_level(logging.DEBUG, logger="instana")
        with tracer.start_as_current_span("test"), patch(
            "instana.instrumentation.redis.collect_attributes",
            side_effect=Exception("test-error"),
        ):
            pipe = self.client.pipeline()
            pipe.set("foox", "barX")
            pipe.set("fooy", "barY")
            pipe.get("foox")
            pipe.execute()
        assert "Error collecting pipeline commands" in caplog.messages

    def test_ignore_redis(
        self,
    ) -> None:
        os.environ["INSTANA_IGNORE_ENDPOINTS"] = "redis"
        agent.options = StandardOptions()

        with tracer.start_as_current_span("test"):
            self.client.set("foox", "barX")
            self.client.get("foox")

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        filtered_spans = agent.filter_spans(spans)
        assert len(filtered_spans) == 1

    def test_ignore_redis_single_command(self) -> None:
        os.environ["INSTANA_IGNORE_ENDPOINTS"] = "redis:set"
        agent.options = StandardOptions()

        with tracer.start_as_current_span("test"):
            self.client.set("foox", "barX")
            self.client.get("foox")

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        filtered_spans = agent.filter_spans(spans)
        assert len(filtered_spans) == 2

        redis_get_span = filtered_spans[0]
        sdk_span = filtered_spans[1]

        assert redis_get_span.n == "redis"
        assert redis_get_span.data["redis"]["command"] == "GET"

        assert sdk_span.n == "sdk"

    def test_ignore_redis_multiple_commands(self) -> None:
        os.environ["INSTANA_IGNORE_ENDPOINTS"] = "redis:set,get"
        agent.options = StandardOptions()
        with tracer.start_as_current_span("test"):
            self.client.set("foox", "barX")
            self.client.get("foox")

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        filtered_spans = agent.filter_spans(spans)
        assert len(filtered_spans) == 1

        sdk_span = filtered_spans[0]

        assert sdk_span.n == "sdk"

    def test_ignore_redis_with_another_instrumentation(self) -> None:
        os.environ["INSTANA_IGNORE_ENDPOINTS"] = "redis:set;something_else:something"
        agent.options = StandardOptions()
        with tracer.start_as_current_span("test"):
            self.client.set("foox", "barX")
            self.client.get("foox")

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        filtered_spans = agent.filter_spans(spans)
        assert len(filtered_spans) == 2

        redis_get_span = filtered_spans[0]
        sdk_span = filtered_spans[1]

        assert redis_get_span.n == "redis"
        assert redis_get_span.data["redis"]["command"] == "GET"

        assert sdk_span.n == "sdk"



# (c) Copyright IBM Corp. 2024

from typing import Generator
from unittest.mock import patch

import pytest

from instana.agent.host import HostAgent
from instana.collector.helpers.resource_usage import ResourceUsage
from instana.collector.helpers.runtime import RuntimeHelper
from instana.collector.host import HostCollector


class TestRuntimeHelper:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.helper = RuntimeHelper(
            collector=HostCollector(
                HostAgent(),
            ),
        )
        yield
        self.helper = None

    def test_default_while_gc_disabled(self) -> None:
        import gc

        gc.disable()
        helper = RuntimeHelper(collector=HostCollector(HostAgent()))
        assert helper.previous_gc_count is None

    def test_collect_metrics(self) -> None:
        response = self.helper.collect_metrics()
        assert response[0]["name"] == "com.instana.plugin.python"

    def test_collect_runtime_snapshot_default(self) -> None:
        plugin_data = self.helper.collect_metrics()
        self.helper._collect_runtime_snapshot(plugin_data[0])
        assert plugin_data[0]["name"] == "com.instana.plugin.python"
        assert plugin_data[0]["data"]["snapshot"]["m"] == "Manual"
        assert len(plugin_data[0]["data"]) == 3

    def test_collect_runtime_snapshot_autowrapt(self) -> None:
        with patch(
            "instana.collector.helpers.runtime.is_autowrapt_instrumented",
            return_value=True,
        ):
            plugin_data = self.helper.collect_metrics()
            self.helper._collect_runtime_snapshot(plugin_data[0])
            assert plugin_data[0]["name"] == "com.instana.plugin.python"
            assert plugin_data[0]["data"]["snapshot"]["m"] == "Autowrapt"
            assert len(plugin_data[0]["data"]) == 3

    def test_collect_runtime_snapshot_webhook(self) -> None:
        with patch(
            "instana.collector.helpers.runtime.is_webhook_instrumented",
            return_value=True,
        ):
            plugin_data = self.helper.collect_metrics()
            self.helper._collect_runtime_snapshot(plugin_data[0])
            assert plugin_data[0]["name"] == "com.instana.plugin.python"
            assert plugin_data[0]["data"]["snapshot"]["m"] == "AutoTrace"
            assert len(plugin_data[0]["data"]) == 3

    def test_collect_gc_metrics(self) -> None:
        plugin_data = self.helper.collect_metrics()

        self.helper._collect_gc_metrics(plugin_data[0], True)
        assert len(self.helper.previous["data"]["metrics"]["gc"]) == 6

    def test_collect_runtime_metrics(self) -> None:
        """Test that _collect_runtime_metrics properly collects metrics"""
        plugin_data = self.helper.collect_metrics()

        # Call the method directly
        self.helper._collect_runtime_metrics(plugin_data[0], True)

        # Verify metrics were collected
        assert "metrics" in plugin_data[0]["data"]
        metrics = plugin_data[0]["data"]["metrics"]

        # Check that resource usage metrics are present
        assert "ru_utime" in metrics
        assert "ru_stime" in metrics
        assert "ru_maxrss" in metrics
        assert "ru_minflt" in metrics
        assert "ru_majflt" in metrics

        # Check that thread metrics are present
        assert "daemon_threads" in metrics
        assert "alive_threads" in metrics
        assert "dummy_threads" in metrics

    def test_runtime_helper_initialization_with_resource_usage(self, mocker):
        """Test that RuntimeHelper initializes with resource_usage"""
        mock_resource = ResourceUsage(
            ru_utime=1.0,
            ru_stime=2.0,
            ru_maxrss=3,
        )
        mocker.patch(
            "instana.collector.helpers.runtime.get_resource_usage",
            return_value=mock_resource,
        )

        helper = RuntimeHelper(collector=HostCollector(HostAgent()))

        assert helper.previous_rusage == mock_resource
        assert helper.previous_rusage.ru_utime == 1.0
        assert helper.previous_rusage.ru_stime == 2.0
        assert helper.previous_rusage.ru_maxrss == 3

    def test_collect_runtime_metrics_with_resource_usage(self, mocker):
        """Test that _collect_runtime_metrics uses resource_usage correctly"""
        # Setup initial state
        initial_resource = ResourceUsage(
            ru_utime=1.0,
            ru_stime=2.0,
            ru_maxrss=3000,
            ru_minflt=100,
            ru_majflt=10,
            ru_nswap=5,
            ru_inblock=200,
            ru_oublock=300,
            ru_msgsnd=10,
            ru_msgrcv=20,
            ru_nsignals=1,
            ru_nvcsw=1000,
            ru_nivcsw=500,
        )
        self.helper.previous_rusage = initial_resource

        # Setup new resource usage values with increments
        new_resource = ResourceUsage(
            ru_utime=1.5,  # +0.5
            ru_stime=3.0,  # +1.0
            ru_maxrss=4000,  # +1000
            ru_minflt=150,  # +50
            ru_majflt=15,  # +5
            ru_nswap=7,  # +2
            ru_inblock=250,  # +50
            ru_oublock=350,  # +50
            ru_msgsnd=15,  # +5
            ru_msgrcv=25,  # +5
            ru_nsignals=3,  # +2
            ru_nvcsw=1200,  # +200
            ru_nivcsw=600,  # +100
        )
        mocker.patch(
            "instana.collector.helpers.runtime.get_resource_usage",
            return_value=new_resource,
        )

        # Call the method
        plugin_data = {"data": {"metrics": {}}}
        self.helper._collect_runtime_metrics(plugin_data, True)

        # Verify metrics were collected with correct deltas
        metrics = plugin_data["data"]["metrics"]
        assert metrics["ru_utime"] == 0.5  # Difference between new and old
        assert metrics["ru_stime"] == 1.0
        assert metrics["ru_maxrss"] == 4000  # This is absolute, not a delta
        assert metrics["ru_minflt"] == 50
        assert metrics["ru_majflt"] == 5
        assert metrics["ru_nswap"] == 2
        assert metrics["ru_inblock"] == 50
        assert metrics["ru_oublock"] == 50
        assert metrics["ru_msgsnd"] == 5
        assert metrics["ru_msgrcv"] == 5
        assert metrics["ru_nsignals"] == 2
        assert metrics["ru_nvcsw"] == 200
        assert metrics["ru_nivcsw"] == 100

        # Verify the previous_rusage was updated
        assert self.helper.previous_rusage == new_resource

    @patch("os.environ")
    def test_collect_runtime_metrics_disabled(self, mock_environ):
        """Test that _collect_runtime_metrics respects INSTANA_DISABLE_METRICS_COLLECTION"""
        # Setup environment variable
        mock_environ.get.return_value = True

        # Call the method
        plugin_data = {"data": {"metrics": {}}}
        self.helper._collect_runtime_metrics(plugin_data, True)

        # Verify no metrics were collected
        assert plugin_data["data"]["metrics"] == {}



# (c) Copyright IBM Corp. 2025


try:
    import contextvars
    import inspect
    from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple

    import kafka  # noqa: F401
    import wrapt
    from opentelemetry import context, trace
    from opentelemetry.trace import SpanKind

    from instana.log import logger
    from instana.propagators.format import Format
    from instana.singletons import get_tracer
    from instana.util.traceutils import (
        get_tracer_tuple,
        tracing_is_off,
    )
    from instana.span.span import InstanaSpan

    if TYPE_CHECKING:
        from kafka.producer.future import FutureRecordMetadata

    consumer_token = None
    consumer_span = contextvars.ContextVar("kafka_python_consumer_span")

    @wrapt.patch_function_wrapper("kafka", "KafkaProducer.send")
    def trace_kafka_send(
        wrapped: Callable[..., "kafka.KafkaProducer.send"],
        instance: "kafka.KafkaProducer",
        args: Tuple[int, str, Tuple[Any, ...]],
        kwargs: Dict[str, Any],
    ) -> "FutureRecordMetadata":
        if tracing_is_off():
            return wrapped(*args, **kwargs)

        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None
        is_suppressed = tracer.exporter._HostAgent__is_endpoint_ignored(
            "kafka",
            "send",
            args[0],
        )
        with tracer.start_as_current_span(
            "kafka-producer", span_context=parent_context, kind=SpanKind.PRODUCER
        ) as span:
            span.set_attribute("kafka.service", args[0])
            span.set_attribute("kafka.access", "send")

            # context propagation
            headers = kwargs.get("headers", [])
            if not is_suppressed and ("x_instana_l_s", b"0") in headers:
                is_suppressed = True

            suppression_header = {"x_instana_l_s": "0" if is_suppressed else "1"}
            headers.append(suppression_header)

            tracer.inject(
                span.context,
                Format.KAFKA_HEADERS,
                headers,
                disable_w3c_trace_context=True,
            )

            headers.remove(suppression_header)

            if tracer.exporter.options.kafka_trace_correlation:
                kwargs["headers"] = headers
            try:
                res = wrapped(*args, **kwargs)
                return res
            except Exception as exc:
                span.record_exception(exc)

    def create_span(
        span_type: str,
        topic: Optional[str],
        headers: Optional[List[Tuple[str, bytes]]] = [],
        exception: Optional[Exception] = None,
    ) -> None:
        try:
            span = consumer_span.get(None)
            if span is not None:
                close_consumer_span(span)

            tracer, parent_span, _ = get_tracer_tuple()

            if not tracer:
                tracer = get_tracer()

            is_suppressed = False
            if topic:
                is_suppressed = tracer.exporter._HostAgent__is_endpoint_ignored(
                    "kafka",
                    span_type,
                    topic,
                )

            if not is_suppressed and headers:
                if ("x_instana_l_s", b"0") in headers:
                    is_suppressed = True

            if is_suppressed:
                return

            parent_context = (
                parent_span.get_span_context()
                if parent_span
                else tracer.extract(
                    Format.KAFKA_HEADERS,
                    headers,
                    disable_w3c_trace_context=True,
                )
            )
            span = tracer.start_span(
                "kafka-consumer", span_context=parent_context, kind=SpanKind.CONSUMER
            )
            if topic:
                span.set_attribute("kafka.service", topic)
            span.set_attribute("kafka.access", span_type)
            if exception:
                span.record_exception(exception)
                span.end()

            save_consumer_span_into_context(span)
        except Exception:
            pass

    def save_consumer_span_into_context(span: "InstanaSpan") -> None:
        global consumer_token
        ctx = trace.set_span_in_context(span)
        consumer_token = context.attach(ctx)
        consumer_span.set(span)

    def close_consumer_span(span: "InstanaSpan") -> None:
        global consumer_token
        if span.is_recording():
            span.end()
            consumer_span.set(None)
        if consumer_token is not None:
            context.detach(consumer_token)
            consumer_token = None

    def clear_context() -> None:
        global consumer_token
        context.attach(trace.set_span_in_context(None))
        consumer_token = None
        consumer_span.set(None)

    @wrapt.patch_function_wrapper("kafka", "KafkaConsumer.__next__")
    def trace_kafka_consume(
        wrapped: Callable[..., "kafka.KafkaConsumer.__next__"],
        instance: "kafka.KafkaConsumer",
        args: Tuple[int, str, Tuple[Any, ...]],
        kwargs: Dict[str, Any],
    ) -> "FutureRecordMetadata":
        exception = None
        res = None

        try:
            res = wrapped(*args, **kwargs)
            create_span(
                "consume",
                res.topic if res else list(instance.subscription())[0],
                res.headers,
            )
            return res
        except StopIteration:
            pass
        except Exception as exc:
            exception = exc
            create_span(
                "consume", list(instance.subscription())[0], exception=exception
            )

    @wrapt.patch_function_wrapper("kafka", "KafkaConsumer.close")
    def trace_kafka_close(
        wrapped: Callable[..., None],
        instance: "kafka.KafkaConsumer",
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
    ) -> None:
        try:
            span = consumer_span.get(None)
            if span is not None:
                close_consumer_span(span)
        except Exception as e:
            logger.debug(
                f"Error while closing kafka-consumer span: {e}"
            )  # pragma: no cover
        return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("kafka", "KafkaConsumer.poll")
    def trace_kafka_poll(
        wrapped: Callable[..., "kafka.KafkaConsumer.poll"],
        instance: "kafka.KafkaConsumer",
        args: Tuple[int, str, Tuple[Any, ...]],
        kwargs: Dict[str, Any],
    ) -> Optional[Dict[str, Any]]:
        # The KafkaConsumer.consume() from the kafka-python-ng call the
        # KafkaConsumer.poll() internally, so we do not consider it here.
        if any(
            frame.function == "trace_kafka_consume"
            for frame in inspect.getouterframes(inspect.currentframe(), 2)
        ):
            return wrapped(*args, **kwargs)

        exception = None
        res = None

        try:
            res = wrapped(*args, **kwargs)
            for partition, consumer_records in res.items():
                for message in consumer_records:
                    create_span(
                        "poll",
                        partition.topic,
                        message.headers if hasattr(message, "headers") else [],
                    )
            return res
        except Exception as exc:
            exception = exc
            create_span("poll", list(instance.subscription())[0], exception=exception)

    logger.debug("Instrumenting Kafka (kafka-python)")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2025
# Assisted by watsonx Code Assistant

import logging
import os
import sys
from typing import TYPE_CHECKING, Generator, List, Union

import pytest

from instana.util.runtime import (
    determine_service_name,
    get_proc_cmdline,
    get_py_source,
    get_runtime_env_info,
    is_ppc64,
    is_s390x,
    is_windows,
    log_runtime_env_info,
)

if TYPE_CHECKING:
    from pytest import LogCaptureFixture
    from pytest_mock import MockerFixture


def test_get_py_source(tmp_path) -> None:
    """Test the get_py_source."""
    filename = "temp_file.py"
    file_contents = "print('Hello, World!')\n"
    expected_output = {"data": file_contents}

    # Create a temporary file for testing purposes.
    temp_file = tmp_path / filename
    temp_file.write_text(file_contents)

    result = get_py_source(f"{tmp_path}/{filename}")
    assert result == expected_output, f"Expected {expected_output}, but got {result}"


@pytest.mark.parametrize(
    "filename, expected_output",
    [
        (
            "non_existent_file.py",
            {"error": "[Errno 2] No such file or directory: 'non_existent_file.py'"},
        ),
        ("temp_file.txt", {"error": "Only Python source files are allowed. (*.py)"}),
    ],
)
def test_get_py_source_error(filename, expected_output) -> None:
    """Test the get_py_source function with various scenarios with errors."""
    result = get_py_source(filename)
    assert result == expected_output, f"Expected {expected_output}, but got {result}"


def test_get_py_source_exception(mocker) -> None:
    """Test the get_py_source function with an exception scenario."""
    exception_message = "No such file or directory"
    mocker.patch(
        "instana.util.runtime.get_py_source", side_effect=Exception(exception_message)
    )

    with pytest.raises(Exception) as exc_info:
        get_py_source("/path/to/non_readable_file.py")
        assert (
            str(exc_info.value) == exception_message
        ), f"Expected {exception_message}, but got {exc_info.value}"


@pytest.fixture()
def _resource_determine_service_name_via_env_var() -> Generator[None, None, None]:
    """SetUp and TearDown"""
    # setup
    yield
    # teardown
    os.environ.pop("INSTANA_SERVICE_NAME", None)
    os.environ.pop("FLASK_APP", None)
    os.environ.pop("DJANGO_SETTINGS_MODULE", None)


@pytest.mark.parametrize(
    "env_var, value, expected_output",
    [
        ("INSTANA_SERVICE_NAME", "test_service", "test_service"),
        ("FLASK_APP", "test_flask_app.py", "test_flask_app.py"),
        ("DJANGO_SETTINGS_MODULE", "test_django_app.settings", "test_django_app"),
    ],
)
def test_determine_service_name_via_env_var(
    env_var: str,
    value: str,
    expected_output: str,
    _resource_determine_service_name_via_env_var: None,
) -> None:
    # Test with multiple environment variables
    os.environ[env_var] = value
    sys.argv = ["something", "nothing"]
    assert determine_service_name() == expected_output


@pytest.mark.parametrize(
    "web_browser, argv, expected_output",
    [
        ("gunicorn", ["gunicorn", "djface.wsgi:app"], "gunicorn"),
        (
            "uwsgi",
            [
                "uwsgi",
                "--master",
                "--processes",
                "4",
                "--threads",
                "2",
                "djface.wsgi:app",
            ],
            "uWSGI master",
        ),
    ],
)
def test_determine_service_name_via_web_browser(
    web_browser: str,
    argv: List[str],
    expected_output: str,
    _resource_determine_service_name_via_env_var: None,
    mocker: "MockerFixture",
) -> None:
    mocker.patch("instana.util.runtime.get_proc_cmdline", return_value="python")
    mocker.patch("os.getpid", return_value=12345)
    sys.argv = argv
    assert determine_service_name() == expected_output


@pytest.mark.parametrize(
    "argv",
    [
        (["python", "test_app.py", "arg1", "arg2"]),
        ([]),
    ],
)
def test_determine_service_name_via_cli_args(
    argv: List[str],
    _resource_determine_service_name_via_env_var: None,
    mocker: "MockerFixture",
) -> None:
    mocker.patch("instana.util.runtime.get_proc_cmdline", return_value="python")
    sys.argv = argv
    # We check "python" in the return of determine_service_name() because this
    # can be the value "python3"
    assert "python" in determine_service_name()


@pytest.mark.parametrize(
    "isatty, expected_output",
    [
        (True, "Interactive Console"),
        (False, ""),
    ],
)
def test_determine_service_name_via_tty(
    isatty: bool,
    expected_output: str,
    _resource_determine_service_name_via_env_var: None,
    mocker: "MockerFixture",
) -> None:
    sys.argv = []
    sys.executable = ""
    sys.stdout.isatty = lambda: isatty
    assert determine_service_name() == expected_output


@pytest.mark.parametrize(
    "as_string, expected",
    [
        (False, ["python", "script.py", "arg1", "arg2"]),
        (True, "python script.py arg1 arg2"),
    ],
)
def test_get_proc_cmdline(
    as_string: bool, expected: Union[List[str], str], mocker: "MockerFixture"
) -> None:
    # Mock the proc filesystem presence
    mocker.patch("os.path.isfile", return_value="/proc/self/cmdline")
    # Mock the content of /proc/self/cmdline
    mocked_data = mocker.mock_open(read_data="python\0script.py\0arg1\0arg2\0")
    mocker.patch("builtins.open", mocked_data)

    assert (
        get_proc_cmdline(as_string) == expected
    ), f"Expected {expected}, but got {get_proc_cmdline(as_string)}"


@pytest.mark.parametrize(
    "as_string, expected",
    [
        (False, ["python"]),
        (True, "python"),
    ],
)
def test_get_proc_cmdline_no_proc_fs(
    as_string: bool, expected: Union[List[str], str], mocker: "MockerFixture"
):
    # Mock the proc filesystem absence
    mocker.patch("os.path.isfile", return_value=False)
    assert get_proc_cmdline(as_string) == expected


def test_get_runtime_env_info(mocker: "MockerFixture") -> None:
    """Test the get_runtime_env_info function."""
    expected_output = ("x86_64", "Linux", "3.13.5")

    mocker.patch("platform.machine", return_value=expected_output[0])
    mocker.patch("platform.system", return_value=expected_output[1])
    mocker.patch("platform.python_version", return_value=expected_output[2])

    machine, system, py_version = get_runtime_env_info()
    assert machine == expected_output[0]
    assert system == expected_output[1]
    assert py_version == expected_output[2]


def test_log_runtime_env_info(
    mocker: "MockerFixture", caplog: "LogCaptureFixture"
) -> None:
    """Test the log_runtime_env_info function."""
    expected_output = ("x86_64", "Linux", "3.13.5")
    caplog.set_level(logging.DEBUG, logger="instana")

    mocker.patch("platform.machine", return_value=expected_output[0])
    mocker.patch("platform.system", return_value=expected_output[1])
    mocker.patch("platform.python_version", return_value=expected_output[2])

    log_runtime_env_info()

    expected_log_message = f"Runtime environment: Machine: {expected_output[0]}, System: {expected_output[1]}, Python version: {expected_output[2]}"
    assert expected_log_message in caplog.messages


@pytest.mark.parametrize(
    "system, expected",
    [
        ("Windows", True),
        ("windows", True),  # Test case insensitivity
        ("WINDOWS", True),  # Test case insensitivity
        ("Linux", False),
        ("Darwin", False),
    ],
)
def test_is_windows(system: str, expected: bool, mocker: "MockerFixture") -> None:
    """Test the is_windows function."""
    mocker.patch(
        "instana.util.runtime.get_runtime_env_info",
        return_value=("x86_64", system, "3.13.5"),
    )
    assert is_windows() == expected


@pytest.mark.parametrize(
    "machine, expected",
    [
        ("ppc64le", True),
        ("ppc64", True),
        ("PPC64", True),  # Test case insensitivity
        ("x86_64", False),
        ("arm64", False),
    ],
)
def test_is_ppc64(machine: str, expected: bool, mocker: "MockerFixture") -> None:
    """Test the is_ppc64 function."""
    mocker.patch(
        "instana.util.runtime.get_runtime_env_info",
        return_value=(machine, "Linux", "3.13.5"),
    )
    assert is_ppc64() == expected


@pytest.mark.parametrize(
    "machine, expected",
    [
        ("s390x", True),
        ("S390X", True),  # Test case insensitivity
        ("x86_64", False),
        ("arm64", False),
    ],
)
def test_is_s390x(machine: str, expected: bool, mocker: "MockerFixture") -> None:
    """Test the is_s390x function."""
    mocker.patch(
        "instana.util.runtime.get_runtime_env_info",
        return_value=(machine, "Linux", "3.13.5"),
    )
    assert is_s390x() == expected



# (c) Copyright IBM Corp. 2025

import logging
from typing import Generator

import pytest
from mock import patch
from opentelemetry.trace.span import format_span_id

from instana.propagators.kafka_propagator import KafkaPropagator
from instana.span_context import SpanContext


class TestKafkaPropagator:
    @pytest.fixture(autouse=True)
    def _resources(self) -> Generator[None, None, None]:
        self.kafka_prop = KafkaPropagator()
        yield

    def test_extract_carrier_headers_as_list_of_dicts(self) -> None:
        carrier_as_a_list = [{"key": "value"}]
        response = self.kafka_prop.extract_carrier_headers(carrier_as_a_list)

        assert response == {"key": "value"}

        carrier_as_a_list = [{"key": "value"}, {"key": "value2"}]
        response = self.kafka_prop.extract_carrier_headers(carrier_as_a_list)

        assert response == {"key": "value2"}

    def test_extract_carrier_headers_as_list_of_tuples(self) -> None:
        carrier_as_a_list = [("key", "value")]
        response = self.kafka_prop.extract_carrier_headers(carrier_as_a_list)

        assert response == {"key": "value"}

        carrier_as_a_list = [("key", "value"), ("key", "value2")]
        response = self.kafka_prop.extract_carrier_headers(carrier_as_a_list)

        assert response == {"key": "value2"}

    def test_extract_carrier_headers_as_dict(self) -> None:
        carrier_as_a_dict = {"key": "value"}
        response = self.kafka_prop.extract_carrier_headers(carrier_as_a_dict)

        assert response == {"key": "value"}

    def test_extract_carrier_headers_as_set(
        self,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.DEBUG, logger="instana")
        carrier_as_a_dict = {"key": "value"}
        with patch.object(
            KafkaPropagator,
            "extract_headers_dict",
            side_effect=Exception(),
        ):
            response = self.kafka_prop.extract_carrier_headers(carrier_as_a_dict)

            assert not response
            assert (
                "kafka_propagator extract_headers_list: Couldn't convert - {'key': 'value'}"
                in caplog.messages
            )

    def test_extract(self) -> None:
        carrier_as_a_dict = {"key": "value"}
        disable_w3c_trace_context = False
        response = self.kafka_prop.extract(carrier_as_a_dict, disable_w3c_trace_context)
        assert response

    def test_extract_with_error(
        self,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.DEBUG, logger="instana")
        carrier_as_a_dict = {"key": "value"}
        disable_w3c_trace_context = False
        with patch.object(
            KafkaPropagator,
            "extract_carrier_headers",
            side_effect=Exception("fake error"),
        ):
            response = self.kafka_prop.extract(
                carrier_as_a_dict, disable_w3c_trace_context
            )
            assert not response
            assert "kafka_propagator extract error: fake error"

    def test_inject_without_suppression(self, trace_id: int, span_id: int) -> None:
        span_context = SpanContext(
            span_id=span_id,
            trace_id=trace_id,
            is_remote=False,
            level=1,
            baggage={},
            sampled=True,
            synthetic=False,
        )
        trace_id = span_context.trace_id
        span_id = span_context.span_id
        carrier = {}

        self.kafka_prop.inject(span_context, carrier)
        assert carrier == {
            "x_instana_l_s": b"1",
            "x_instana_t": format_span_id(trace_id).encode("utf-8"),
            "x_instana_s": format_span_id(span_id).encode("utf-8"),
        }

    def test_inject_with_suppression(self, trace_id: int, span_id: int) -> None:
        span_context = SpanContext(
            span_id=span_id,
            trace_id=trace_id,
            is_remote=False,
            level=1,
            baggage={},
            sampled=True,
            synthetic=False,
        )
        trace_id = span_context.trace_id
        span_id = span_context.span_id
        carrier = {"x_instana_l_s": "0"}

        self.kafka_prop.inject(span_context, carrier)
        assert carrier == {"x_instana_l_s": b"0"}



# (c) Copyright IBM Corp. 2025

import logging
import os
from typing import Generator

import pytest
from mock import patch

from instana.configurator import config
from instana.options import (
    AWSFargateOptions,
    AWSLambdaOptions,
    BaseOptions,
    EKSFargateOptions,
    GCROptions,
    ServerlessOptions,
    StandardOptions,
)


class TestBaseOptions:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.base_options = None
        yield
        if "tracing" in config.keys():
            del config["tracing"]

    def test_base_options(self) -> None:
        if "INSTANA_DEBUG" in os.environ:
            del os.environ["INSTANA_DEBUG"]
        self.base_options = BaseOptions()

        assert not self.base_options.debug
        assert self.base_options.log_level == logging.WARN
        assert not self.base_options.extra_http_headers
        assert not self.base_options.allow_exit_as_root
        assert not self.base_options.ignore_endpoints
        assert self.base_options.kafka_trace_correlation
        assert self.base_options.secrets_matcher == "contains-ignore-case"
        assert self.base_options.secrets_list == ["key", "pass", "secret"]
        assert not self.base_options.secrets
        assert self.base_options.disabled_spans == []
        assert self.base_options.enabled_spans == []

    def test_base_options_with_config(self) -> None:
        config["tracing"] = {
            "ignore_endpoints": "service1;service3:method1,method2",
            "kafka": {"trace_correlation": True},
        }
        self.base_options = BaseOptions()
        assert self.base_options.ignore_endpoints == [
            "service1.*",
            "service3.method1",
            "service3.method2",
        ]
        assert self.base_options.kafka_trace_correlation

    @patch.dict(
        os.environ,
        {
            "INSTANA_DEBUG": "true",
            "INSTANA_EXTRA_HTTP_HEADERS": "SOMETHING;HERE",
            "INSTANA_IGNORE_ENDPOINTS": "service1;service2:method1,method2",
            "INSTANA_SECRETS": "secret1:username,password",
            "INSTANA_TRACING_DISABLE": "logging, redis,kafka",
        },
    )
    def test_base_options_with_env_vars(self) -> None:
        self.base_options = BaseOptions()
        assert self.base_options.log_level == logging.DEBUG
        assert self.base_options.debug

        assert self.base_options.extra_http_headers == ["something", "here"]

        assert self.base_options.ignore_endpoints == [
            "service1.*",
            "service2.method1",
            "service2.method2",
        ]

        assert self.base_options.secrets_matcher == "secret1"
        assert self.base_options.secrets_list == ["username", "password"]

        assert "logging" in self.base_options.disabled_spans
        assert "redis" in self.base_options.disabled_spans
        assert "kafka" in self.base_options.disabled_spans
        assert len(self.base_options.enabled_spans) == 0

    @patch.dict(
        os.environ,
        {"INSTANA_IGNORE_ENDPOINTS_PATH": "tests/util/test_configuration-1.yaml"},
    )
    def test_base_options_with_endpoint_file(self) -> None:
        self.base_options = BaseOptions()
        assert self.base_options.ignore_endpoints == [
            "redis.get",
            "redis.type",
            "dynamodb.query",
            "kafka.consume.span-topic",
            "kafka.consume.topic1",
            "kafka.consume.topic2",
            "kafka.send.span-topic",
            "kafka.send.topic1",
            "kafka.send.topic2",
            "kafka.consume.topic3",
            "kafka.*.span-topic",
            "kafka.*.topic4",
        ]
        del self.base_options

    @patch.dict(
        os.environ,
        {
            "INSTANA_IGNORE_ENDPOINTS": "env_service1;env_service2:method1,method2",
            "INSTANA_KAFKA_TRACE_CORRELATION": "false",
            "INSTANA_IGNORE_ENDPOINTS_PATH": "tests/util/test_configuration-1.yaml",
            "INSTANA_TRACING_DISABLE": "logging,redis, kafka",
        },
    )
    def test_set_trace_configurations_by_env_variable(self) -> None:
        # The priority is as follows:
        # environment variables > in-code configuration >
        # > agent config (configuration.yaml) > default value

        # in-code configuration
        config["tracing"] = {}
        config["tracing"]["ignore_endpoints"] = (
            "config_service1;config_service2:method1,method2"
        )
        config["tracing"]["kafka"] = {"trace_correlation": True}
        config["tracing"]["disable"] = [{"databases": True}]

        # agent config (configuration.yaml)
        test_tracing = {
            "ignore-endpoints": "service1;service2:method1,method2",
            "disable": [
                {"messaging": True},
            ],
        }

        # Setting by env variable
        self.base_options = StandardOptions()
        self.base_options.set_tracing(test_tracing)

        assert self.base_options.ignore_endpoints == [
            "env_service1.*",
            "env_service2.method1",
            "env_service2.method2",
        ]
        assert not self.base_options.kafka_trace_correlation

        # Check disabled_spans list
        assert "logging" in self.base_options.disabled_spans
        assert "redis" in self.base_options.disabled_spans
        assert "kafka" in self.base_options.disabled_spans
        assert "databases" not in self.base_options.disabled_spans
        assert "messaging" not in self.base_options.disabled_spans
        assert len(self.base_options.enabled_spans) == 0

    @patch.dict(
        os.environ,
        {
            "INSTANA_KAFKA_TRACE_CORRELATION": "false",
            "INSTANA_IGNORE_ENDPOINTS_PATH": "tests/util/test_configuration-1.yaml",
        },
    )
    def test_set_trace_configurations_by_in_code_configuration(self) -> None:
        # The priority is as follows:
        # in-code configuration > agent config (configuration.yaml) > default value

        # in-code configuration
        config["tracing"] = {}
        config["tracing"]["ignore_endpoints"] = (
            "config_service1;config_service2:method1,method2"
        )
        config["tracing"]["kafka"] = {"trace_correlation": True}
        config["tracing"]["disable"] = [{"databases": True}]

        # agent config (configuration.yaml)
        test_tracing = {
            "ignore-endpoints": "service1;service2:method1,method2",
            "disable": [
                {"messaging": True},
            ],
        }

        self.base_options = StandardOptions()
        self.base_options.set_tracing(test_tracing)

        assert self.base_options.ignore_endpoints == [
            "redis.get",
            "redis.type",
            "dynamodb.query",
            "kafka.consume.span-topic",
            "kafka.consume.topic1",
            "kafka.consume.topic2",
            "kafka.send.span-topic",
            "kafka.send.topic1",
            "kafka.send.topic2",
            "kafka.consume.topic3",
            "kafka.*.span-topic",
            "kafka.*.topic4",
        ]

        # Check disabled_spans list
        assert "databases" in self.base_options.disabled_spans
        assert "logging" not in self.base_options.disabled_spans
        assert "redis" not in self.base_options.disabled_spans
        assert "kafka" not in self.base_options.disabled_spans
        assert "messaging" not in self.base_options.disabled_spans
        assert len(self.base_options.enabled_spans) == 0

    def test_set_trace_configurations_by_in_code_variable(self) -> None:
        config["tracing"] = {}
        config["tracing"]["ignore_endpoints"] = (
            "config_service1;config_service2:method1,method2"
        )
        config["tracing"]["kafka"] = {"trace_correlation": True}
        test_tracing = {"ignore-endpoints": "service1;service2:method1,method2"}

        self.base_options = StandardOptions()
        self.base_options.set_tracing(test_tracing)

        assert self.base_options.ignore_endpoints == [
            "config_service1.*",
            "config_service2.method1",
            "config_service2.method2",
        ]
        assert self.base_options.kafka_trace_correlation

    def test_set_trace_configurations_by_agent_configuration(self) -> None:
        test_tracing = {
            "ignore-endpoints": "service1;service2:method1,method2",
            "trace-correlation": True,
            "disable": [
                {
                    "messaging": True,
                    "logging": True,
                    "kafka": False,
                },
            ],
        }

        self.base_options = StandardOptions()
        self.base_options.set_tracing(test_tracing)

        assert self.base_options.ignore_endpoints == [
            "service1.*",
            "service2.method1",
            "service2.method2",
        ]
        assert self.base_options.kafka_trace_correlation

        # Check disabled_spans list
        assert "databases" not in self.base_options.disabled_spans
        assert "logging" in self.base_options.disabled_spans
        assert "messaging" in self.base_options.disabled_spans
        assert "kafka" in self.base_options.enabled_spans

    def test_set_trace_configurations_by_default(self) -> None:
        self.base_options = StandardOptions()
        self.base_options.set_tracing({})

        assert not self.base_options.ignore_endpoints
        assert self.base_options.kafka_trace_correlation
        assert len(self.base_options.disabled_spans) == 0
        assert len(self.base_options.enabled_spans) == 0

    @patch.dict(
        os.environ,
        {"INSTANA_TRACING_DISABLE": "true"},
    )
    def test_set_trace_configurations_disable_all_tracing(self) -> None:
        self.base_options = BaseOptions()

        # All categories should be disabled
        assert "logging" in self.base_options.disabled_spans
        assert "databases" in self.base_options.disabled_spans
        assert "messaging" in self.base_options.disabled_spans
        assert "protocols" in self.base_options.disabled_spans

        # Check is_span_disabled method
        assert self.base_options.is_span_disabled(category="logging")
        assert self.base_options.is_span_disabled(category="databases")
        assert self.base_options.is_span_disabled(span_type="redis")

    @patch.dict(
        os.environ,
        {
            "INSTANA_CONFIG_PATH": "tests/util/test_configuration-1.yaml",
        },
    )
    def test_set_trace_configurations_disable_local_yaml(self) -> None:
        self.base_options = BaseOptions()

        # All categories should be disabled
        assert "logging" in self.base_options.disabled_spans
        assert "databases" in self.base_options.disabled_spans
        assert "redis" not in self.base_options.disabled_spans
        assert "redis" in self.base_options.enabled_spans

        # Check is_span_disabled method
        assert self.base_options.is_span_disabled(category="logging")
        assert self.base_options.is_span_disabled(category="databases")
        assert not self.base_options.is_span_disabled(span_type="redis")

    def test_is_span_disabled_method(self) -> None:
        self.base_options = BaseOptions()

        # Default behavior - nothing disabled
        assert not self.base_options.is_span_disabled(category="logging")
        assert not self.base_options.is_span_disabled(span_type="redis")

        # Disable a category
        self.base_options.disabled_spans = ["databases"]
        assert not self.base_options.is_span_disabled(category="logging")
        assert self.base_options.is_span_disabled(category="databases")
        assert self.base_options.is_span_disabled(span_type="redis")
        assert self.base_options.is_span_disabled(span_type="mysql")

        # Test precedence rules
        self.base_options.enabled_spans = ["redis"]
        assert self.base_options.is_span_disabled(category="databases")
        assert self.base_options.is_span_disabled(span_type="mysql")
        assert not self.base_options.is_span_disabled(span_type="redis")


class TestStandardOptions:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.standart_options = None
        yield
        if "tracing" in config.keys():
            del config["tracing"]

    def test_standard_options(self) -> None:
        self.standart_options = StandardOptions()

        assert self.standart_options.AGENT_DEFAULT_HOST == "localhost"
        assert self.standart_options.AGENT_DEFAULT_PORT == 42699

    def test_set_secrets(self) -> None:
        self.standart_options = StandardOptions()

        test_secrets = {"matcher": "sample-match", "list": ["sample", "list"]}
        self.standart_options.set_secrets(test_secrets)
        assert self.standart_options.secrets_matcher == "sample-match"
        assert self.standart_options.secrets_list == ["sample", "list"]

    def test_set_extra_headers(self) -> None:
        self.standart_options = StandardOptions()
        test_headers = {"header1": "sample-match", "header2": ["sample", "list"]}

        self.standart_options.set_extra_headers(test_headers)
        assert self.standart_options.extra_http_headers == test_headers

    def test_set_tracing(
        self,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.DEBUG, logger="instana")
        self.standart_options = StandardOptions()

        test_tracing = {
            "ignore-endpoints": "service1;service2:method1,method2",
            "kafka": {"trace-correlation": "false", "header-format": "binary"},
        }
        self.standart_options.set_tracing(test_tracing)

        assert self.standart_options.ignore_endpoints == [
            "service1.*",
            "service2.method1",
            "service2.method2",
        ]
        assert not self.standart_options.kafka_trace_correlation
        assert (
            "Binary header format for Kafka is deprecated. Please use string header format."
            in caplog.messages
        )
        assert not self.standart_options.extra_http_headers

    def test_set_tracing_with_span_disabling(self) -> None:
        self.standart_options = StandardOptions()

        test_tracing = {
            "disable": [{"logging": True}, {"redis": False}, {"databases": True}]
        }
        self.standart_options.set_tracing(test_tracing)

        # Check disabled_spans and enabled_spans lists
        assert "logging" in self.standart_options.disabled_spans
        assert "databases" in self.standart_options.disabled_spans
        assert "redis" in self.standart_options.enabled_spans

        # Check is_span_disabled method
        assert self.standart_options.is_span_disabled(category="logging")
        assert self.standart_options.is_span_disabled(category="databases")
        assert self.standart_options.is_span_disabled(span_type="mysql")
        assert not self.standart_options.is_span_disabled(span_type="redis")

    def test_set_from(self) -> None:
        self.standart_options = StandardOptions()
        test_res_data = {
            "secrets": {"matcher": "sample-match", "list": ["sample", "list"]},
            "tracing": {"ignore-endpoints": "service1;service2:method1,method2"},
        }
        self.standart_options.set_from(test_res_data)

        assert (
            self.standart_options.secrets_matcher == test_res_data["secrets"]["matcher"]
        )
        assert self.standart_options.secrets_list == test_res_data["secrets"]["list"]
        assert self.standart_options.ignore_endpoints == [
            "service1.*",
            "service2.method1",
            "service2.method2",
        ]

        test_res_data = {
            "extraHeaders": {"header1": "sample-match", "header2": ["sample", "list"]},
        }
        self.standart_options.set_from(test_res_data)

        assert self.standart_options.extra_http_headers == test_res_data["extraHeaders"]

    def test_set_from_bool(
        self,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.DEBUG, logger="instana")
        caplog.clear()

        self.standart_options = StandardOptions()
        test_res_data = True
        self.standart_options.set_from(test_res_data)

        assert len(caplog.messages) == 1
        assert len(caplog.records) == 1
        assert (
            "options.set_from: Wrong data type - <class 'bool'>" in caplog.messages[0]
        )

        assert self.standart_options.secrets_list == ["key", "pass", "secret"]
        assert self.standart_options.ignore_endpoints == []
        assert not self.standart_options.extra_http_headers


class TestServerlessOptions:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.serverless_options = None
        yield

    def test_serverless_options(self) -> None:
        self.serverless_options = ServerlessOptions()

        assert not self.serverless_options.debug
        assert self.serverless_options.log_level == logging.WARN
        assert not self.serverless_options.extra_http_headers
        assert not self.serverless_options.allow_exit_as_root
        assert not self.serverless_options.ignore_endpoints
        assert self.serverless_options.secrets_matcher == "contains-ignore-case"
        assert self.serverless_options.secrets_list == ["key", "pass", "secret"]
        assert not self.serverless_options.secrets
        assert not self.serverless_options.agent_key
        assert not self.serverless_options.endpoint_url
        assert self.serverless_options.ssl_verify
        assert not self.serverless_options.endpoint_proxy
        assert self.serverless_options.timeout == 0.8

    @patch.dict(
        os.environ,
        {
            "INSTANA_AGENT_KEY": "key1",
            "INSTANA_ENDPOINT_URL": "localhost",
            "INSTANA_DISABLE_CA_CHECK": "true",
            "INSTANA_ENDPOINT_PROXY": "proxy1",
            "INSTANA_TIMEOUT": "3000",
            "INSTANA_LOG_LEVEL": "info",
        },
    )
    def test_serverless_options_with_env_vars(self) -> None:
        self.serverless_options = ServerlessOptions()

        assert self.serverless_options.agent_key == "key1"
        assert self.serverless_options.endpoint_url == "localhost"
        assert not self.serverless_options.ssl_verify
        assert self.serverless_options.endpoint_proxy == {"https": "proxy1"}
        assert self.serverless_options.timeout == 3
        assert self.serverless_options.log_level == logging.INFO


class TestAWSLambdaOptions:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.aws_lambda_options = None
        yield

    def test_aws_lambda_options(self) -> None:
        self.aws_lambda_options = AWSLambdaOptions()

        assert not self.aws_lambda_options.agent_key
        assert not self.aws_lambda_options.endpoint_url
        assert self.aws_lambda_options.ssl_verify
        assert not self.aws_lambda_options.endpoint_proxy
        assert self.aws_lambda_options.timeout == 0.8
        assert self.aws_lambda_options.log_level == logging.WARN


class TestAWSFargateOptions:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.aws_fargate_options = None
        yield

    def test_aws_fargate_options(self) -> None:
        self.aws_fargate_options = AWSFargateOptions()

        assert not self.aws_fargate_options.agent_key
        assert not self.aws_fargate_options.endpoint_url
        assert self.aws_fargate_options.ssl_verify
        assert not self.aws_fargate_options.endpoint_proxy
        assert self.aws_fargate_options.timeout == 0.8
        assert self.aws_fargate_options.log_level == logging.WARN
        assert not self.aws_fargate_options.tags
        assert not self.aws_fargate_options.zone

    @patch.dict(
        os.environ,
        {
            "INSTANA_AGENT_KEY": "key1",
            "INSTANA_ENDPOINT_URL": "localhost",
            "INSTANA_DISABLE_CA_CHECK": "true",
            "INSTANA_ENDPOINT_PROXY": "proxy1",
            "INSTANA_TIMEOUT": "3000",
            "INSTANA_LOG_LEVEL": "info",
            "INSTANA_TAGS": "key1=value1,key2=value2",
            "INSTANA_ZONE": "zone1",
        },
    )
    def test_aws_fargate_options_with_env_vars(self) -> None:
        self.aws_fargate_options = AWSFargateOptions()

        assert self.aws_fargate_options.agent_key == "key1"
        assert self.aws_fargate_options.endpoint_url == "localhost"
        assert not self.aws_fargate_options.ssl_verify
        assert self.aws_fargate_options.endpoint_proxy == {"https": "proxy1"}
        assert self.aws_fargate_options.timeout == 3
        assert self.aws_fargate_options.log_level == logging.INFO

        assert self.aws_fargate_options.tags == {"key1": "value1", "key2": "value2"}
        assert self.aws_fargate_options.zone == "zone1"


class TestEKSFargateOptions:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.eks_fargate_options = None
        yield

    def test_eks_fargate_options(self) -> None:
        self.eks_fargate_options = EKSFargateOptions()

        assert not self.eks_fargate_options.agent_key
        assert not self.eks_fargate_options.endpoint_url
        assert self.eks_fargate_options.ssl_verify
        assert not self.eks_fargate_options.endpoint_proxy
        assert self.eks_fargate_options.timeout == 0.8
        assert self.eks_fargate_options.log_level == logging.WARN

    @patch.dict(
        os.environ,
        {
            "INSTANA_AGENT_KEY": "key1",
            "INSTANA_ENDPOINT_URL": "localhost",
            "INSTANA_DISABLE_CA_CHECK": "true",
            "INSTANA_ENDPOINT_PROXY": "proxy1",
            "INSTANA_TIMEOUT": "3000",
            "INSTANA_LOG_LEVEL": "info",
        },
    )
    def test_eks_fargate_options_with_env_vars(self) -> None:
        self.eks_fargate_options = EKSFargateOptions()

        assert self.eks_fargate_options.agent_key == "key1"
        assert self.eks_fargate_options.endpoint_url == "localhost"
        assert not self.eks_fargate_options.ssl_verify
        assert self.eks_fargate_options.endpoint_proxy == {"https": "proxy1"}
        assert self.eks_fargate_options.timeout == 3
        assert self.eks_fargate_options.log_level == logging.INFO


class TestGCROptions:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.gcr_options = None
        yield

    def test_gcr_options(self) -> None:
        self.gcr_options = GCROptions()

        assert not self.gcr_options.debug
        assert self.gcr_options.log_level == logging.WARN
        assert not self.gcr_options.extra_http_headers
        assert not self.gcr_options.allow_exit_as_root
        assert not self.gcr_options.ignore_endpoints
        assert self.gcr_options.secrets_matcher == "contains-ignore-case"
        assert self.gcr_options.secrets_list == ["key", "pass", "secret"]
        assert not self.gcr_options.secrets
        assert not self.gcr_options.agent_key
        assert not self.gcr_options.endpoint_url
        assert self.gcr_options.ssl_verify
        assert not self.gcr_options.endpoint_proxy
        assert self.gcr_options.timeout == 0.8

    @patch.dict(
        os.environ,
        {
            "INSTANA_AGENT_KEY": "key1",
            "INSTANA_ENDPOINT_URL": "localhost",
            "INSTANA_DISABLE_CA_CHECK": "true",
            "INSTANA_ENDPOINT_PROXY": "proxy1",
            "INSTANA_TIMEOUT": "3000",
            "INSTANA_LOG_LEVEL": "info",
        },
    )
    def test_gcr_options_with_env_vars(self) -> None:
        self.gcr_options = GCROptions()

        assert self.gcr_options.agent_key == "key1"
        assert self.gcr_options.endpoint_url == "localhost"
        assert not self.gcr_options.ssl_verify
        assert self.gcr_options.endpoint_proxy == {"https": "proxy1"}
        assert self.gcr_options.timeout == 3
        assert self.gcr_options.log_level == logging.INFO


# Made with Bob



# (c) Copyright IBM Corp. 2025

try:
    import spyne
    import wrapt
    from typing import TYPE_CHECKING, Dict, Any, Callable, Tuple, Iterable, Type, Optional

    from types import SimpleNamespace

    from instana.log import logger
    from instana.singletons import agent, tracer
    from instana.propagators.format import Format
    from instana.util.secrets import strip_secrets_from_query

    if TYPE_CHECKING:
        from instana.span.span import InstanaSpan
        from spyne.application import Application
        from spyne.server.wsgi import WsgiApplication

    def set_span_attributes(span: "InstanaSpan", headers: Dict[str, Any]) -> None:
        if "PATH_INFO" in headers:
            span.set_attribute("rpc.call", headers["PATH_INFO"])
        if "QUERY_STRING" in headers and len(headers["QUERY_STRING"]):
            scrubbed_params = strip_secrets_from_query(
                headers["QUERY_STRING"],
                agent.options.secrets_matcher,
                agent.options.secrets_list,
            )
            span.set_attribute("rpc.params", scrubbed_params)
        if "REMOTE_ADDR" in headers:
            span.set_attribute("rpc.host", headers["REMOTE_ADDR"])
        if "SERVER_PORT" in headers:
            span.set_attribute("rpc.port", headers["SERVER_PORT"])

    def record_error(span: "InstanaSpan", response_string: str, error: Optional[Type[Exception]]) -> None:
        resp_code = int(response_string.split()[0])

        if 500 <= resp_code:
            span.record_exception(error)


    @wrapt.patch_function_wrapper("spyne.server.wsgi", "WsgiApplication.handle_error")
    def handle_error_with_instana(
        wrapped: Callable[..., Iterable[object]],
        instance: "WsgiApplication",
        args: Tuple[object],
        kwargs: Dict[str, Any],
    ) -> Iterable[object]:
        ctx = args[0]

        # span created inside process_request() will be handled by finalize() method
        if ctx.udc and ctx.udc.span:
            return wrapped(*args, **kwargs)

        headers = ctx.transport.req_env
        span_context = tracer.extract(Format.HTTP_HEADERS, headers)

        with tracer.start_as_current_span("rpc-server", span_context=span_context) as span:
            set_span_attributes(span, headers)

            response_headers = ctx.transport.resp_headers

            tracer.inject(span.context, Format.HTTP_HEADERS, response_headers)

            response = wrapped(*args, **kwargs)

            record_error(span, ctx.transport.resp_code, ctx.in_error or ctx.out_error)
            return response

    @wrapt.patch_function_wrapper(
        "spyne.server.wsgi", "WsgiApplication._WsgiApplication__finalize"
    )
    def finalize_with_instana(
        wrapped: Callable[..., Tuple[()]],
        instance: "WsgiApplication",
        args: Tuple[object],
        kwargs: Dict[str, Any],
    ) -> Tuple[()]:
        ctx = args[0]
        response_string = ctx.transport.resp_code

        if ctx.udc and ctx.udc.span and response_string:
            span = ctx.udc.span
            record_error(span, response_string, ctx.in_error or ctx.out_error)
            if span.is_recording():
                span.end()

            ctx.udc.span = None
        return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("spyne.application", "Application.process_request")
    def process_request_with_instana(
        wrapped: Callable[..., None],
        instance: "Application",
        args: Tuple[object],
        kwargs: Dict[str, Any],
    ) -> None:
        ctx = args[0]
        headers = ctx.transport.req_env
        span_context = tracer.extract(Format.HTTP_HEADERS, headers)

        with tracer.start_as_current_span(
            "rpc-server",
            span_context=span_context,
            end_on_exit=False,
        ) as span:
            set_span_attributes(span, headers)

            response = wrapped(*args, **kwargs)
            response_headers = ctx.transport.resp_headers

            tracer.inject(span.context, Format.HTTP_HEADERS, response_headers)

            ## Store the span in the user defined context object offered by Spyne
            if ctx.udc:
                ctx.udc.span = span
            else:
                ctx.udc = SimpleNamespace()
                ctx.udc.span = span
            return response

    logger.debug("Instrumenting Spyne")

except ImportError:
    pass



# (c) Copyright IBM Corp. 2025

try:
    import aio_pika
    import wrapt
    from typing import (
        TYPE_CHECKING,
        Dict,
        Any,
        Callable,
        Tuple,
        Type,
        Optional,
    )

    from instana.log import logger
    from instana.propagators.format import Format
    from instana.util.traceutils import get_tracer_tuple, tracing_is_off
    from instana.singletons import tracer

    if TYPE_CHECKING:
        from instana.span.span import InstanaSpan
        from aio_pika.exchange import Exchange
        from aiormq.abc import ConfirmationFrameType
        from aio_pika.abc import ConsumerTag, AbstractMessage
        from aio_pika.queue import Queue, QueueIterator

    def _extract_span_attributes(
        span: "InstanaSpan", connection, sort: str, routing_key: str, exchange: str
    ) -> None:
        span.set_attribute("address", str(connection.url))

        span.set_attribute("sort", sort)
        span.set_attribute("key", routing_key)
        span.set_attribute("exchange", exchange)

    @wrapt.patch_function_wrapper("aio_pika", "Exchange.publish")
    async def publish_with_instana(
        wrapped: Callable[..., Optional["ConfirmationFrameType"]],
        instance: "Exchange",
        args: Tuple[object],
        kwargs: Dict[str, Any],
    ) -> Optional["ConfirmationFrameType"]:
        if tracing_is_off():
            return await wrapped(*args, **kwargs)

        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None

        def _bind_args(
            message: Type["AbstractMessage"],
            routing_key: str,
            *args: object,
            **kwargs: object,
        ) -> Tuple[object, ...]:
            return (message, routing_key, args, kwargs)
        
        (message, routing_key, args, kwargs) = _bind_args(
            *args, **kwargs
        )

        with tracer.start_as_current_span(
            "rabbitmq", span_context=parent_context
        ) as span:
            connection = instance.channel._connection

            _extract_span_attributes(
                span, connection, "publish", routing_key, instance.name
            )

            tracer.inject(
                span.context,
                Format.HTTP_HEADERS,
                message.properties.headers,
                disable_w3c_trace_context=True,
            )

            args = (message, routing_key) + args

            try:
                response = await wrapped(*args, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
            else:
                return response

    @wrapt.patch_function_wrapper("aio_pika", "Queue.consume")
    async def consume_with_instana(
        wrapped: Callable[..., "ConsumerTag"],
        instance: Type["Queue"],
        args: Tuple[object],
        kwargs: Dict[str, Any],
    ) -> "ConsumerTag":
        connection = instance.channel._connection
        callback = kwargs["callback"] if kwargs.get("callback") else args[0]

        @wrapt.decorator
        async def callback_wrapper(
            wrapped: Callable[[Type["AbstractMessage"]], Any],
            instance: Type["QueueIterator"],
            args: Tuple[Type["AbstractMessage"], ...],
            kwargs: Dict[str, Any],
        ) -> Callable[[Type["AbstractMessage"]], Any]:
            message = args[0]
            parent_context = tracer.extract(
                Format.HTTP_HEADERS, message.headers, disable_w3c_trace_context=True
            )
            with tracer.start_as_current_span(
                "rabbitmq", span_context=parent_context
            ) as span:
                _extract_span_attributes(
                    span, connection, "consume", message.routing_key, message.exchange
                )
                try:
                    response = await wrapped(*args, **kwargs)
                except Exception as exc:
                    span.record_exception(exc)
                else:
                    return response

        wrapped_callback = callback_wrapper(callback)
        if kwargs.get("callback"):
            kwargs["callback"] = wrapped_callback
        else:
            args = (wrapped_callback,) + args[1:]

        return await wrapped(*args, **kwargs)

    logger.debug("Instrumenting aio-pika")

except ImportError:
    pass



# (c) Copyright IBM Corp. 2025

import pytest

from instana.options import BaseOptions, StandardOptions
from instana.singletons import agent


class TestSpanDisabling:
    @pytest.fixture(autouse=True)
    def setup(self):
        # Save original options
        self.original_options = agent.options
        yield
        # Restore original options
        agent.options = self.original_options

    def test_is_span_disabled_default(self):
        options = BaseOptions()
        assert not options.is_span_disabled(category="logging")
        assert not options.is_span_disabled(category="databases")
        assert not options.is_span_disabled(span_type="redis")

    def test_disable_category(self):
        options = BaseOptions()
        options.disabled_spans = ["logging"]
        assert options.is_span_disabled(category="logging")
        assert not options.is_span_disabled(category="databases")

    def test_disable_type(self):
        options = BaseOptions()
        options.disabled_spans = ["redis"]
        assert options.is_span_disabled(span_type="redis")
        assert not options.is_span_disabled(span_type="mysql")

    def test_type_category_relationship(self):
        options = BaseOptions()
        options.disabled_spans = ["databases"]
        assert options.is_span_disabled(span_type="redis")
        assert options.is_span_disabled(span_type="mysql")

    def test_precedence_rules(self):
        options = BaseOptions()
        options.disabled_spans = ["databases"]
        options.enabled_spans = ["redis"]
        assert options.is_span_disabled(category="databases")
        assert options.is_span_disabled(span_type="mysql")
        assert not options.is_span_disabled(span_type="redis")

    @pytest.mark.parametrize("value", ["True", "true", "1"])
    def test_env_var_disable_all(self, value, monkeypatch):
        monkeypatch.setenv("INSTANA_TRACING_DISABLE", value)
        options = BaseOptions()
        assert options.is_span_disabled(category="logging") is True
        assert options.is_span_disabled(category="databases") is True
        assert options.is_span_disabled(category="messaging") is True
        assert options.is_span_disabled(category="protocols") is True

    def test_env_var_disable_specific(self, monkeypatch):
        monkeypatch.setenv("INSTANA_TRACING_DISABLE", "logging, redis")
        options = BaseOptions()
        assert options.is_span_disabled(category="logging") is True
        assert options.is_span_disabled(category="databases") is False
        assert options.is_span_disabled(span_type="redis") is True
        assert options.is_span_disabled(span_type="mysql") is False

    def test_yaml_config(self):
        options = StandardOptions()
        tracing_config = {
            "disable": [{"logging": True}, {"redis": False}, {"databases": True}]
        }
        options.set_tracing(tracing_config)
        assert options.is_span_disabled(category="logging")
        assert options.is_span_disabled(category="databases")
        assert options.is_span_disabled(span_type="mysql")
        assert not options.is_span_disabled(span_type="redis")


# Made with Bob



# (c) Copyright IBM Corp. 2025


import pytest

from instana.collector.helpers.resource_usage import (
    ResourceUsage,
    _get_unix_resource_usage,
    _get_windows_resource_usage,
    get_resource_usage,
)
from instana.util.runtime import is_windows


class TestResourceUsage:
    def test_resource_usage_namedtuple_defaults(self):
        """Test that ResourceUsage has proper default values"""
        usage = ResourceUsage()
        assert usage.ru_utime == 0.0
        assert usage.ru_stime == 0.0
        assert usage.ru_maxrss == 0
        assert usage.ru_ixrss == 0
        assert usage.ru_idrss == 0
        assert usage.ru_isrss == 0
        assert usage.ru_minflt == 0
        assert usage.ru_majflt == 0
        assert usage.ru_nswap == 0
        assert usage.ru_inblock == 0
        assert usage.ru_oublock == 0
        assert usage.ru_msgsnd == 0
        assert usage.ru_msgrcv == 0
        assert usage.ru_nsignals == 0
        assert usage.ru_nvcsw == 0
        assert usage.ru_nivcsw == 0

    def test_resource_usage_namedtuple_custom_values(self):
        """Test that ResourceUsage can be initialized with custom values"""
        usage = ResourceUsage(
            ru_utime=1.0,
            ru_stime=2.0,
            ru_maxrss=3,
            ru_ixrss=4,
            ru_idrss=5,
            ru_isrss=6,
            ru_minflt=7,
            ru_majflt=8,
            ru_nswap=9,
            ru_inblock=10,
            ru_oublock=11,
            ru_msgsnd=12,
            ru_msgrcv=13,
            ru_nsignals=14,
            ru_nvcsw=15,
            ru_nivcsw=16,
        )
        assert usage.ru_utime == 1.0
        assert usage.ru_stime == 2.0
        assert usage.ru_maxrss == 3
        assert usage.ru_ixrss == 4
        assert usage.ru_idrss == 5
        assert usage.ru_isrss == 6
        assert usage.ru_minflt == 7
        assert usage.ru_majflt == 8
        assert usage.ru_nswap == 9
        assert usage.ru_inblock == 10
        assert usage.ru_oublock == 11
        assert usage.ru_msgsnd == 12
        assert usage.ru_msgrcv == 13
        assert usage.ru_nsignals == 14
        assert usage.ru_nvcsw == 15
        assert usage.ru_nivcsw == 16

    @pytest.mark.skipif(
        is_windows(),
        reason="Avoiding Unix resource usage collection on Windows systems.",
    )
    def test_get_resource_usage_unix(self):
        """Test that get_resource_usage calls _get_unix_resource_usage on Unix-like systems."""
        usage = get_resource_usage()

        assert usage.ru_utime >= 0.0
        assert usage.ru_stime >= 0.0
        assert usage.ru_maxrss >= 0
        assert usage.ru_ixrss >= 0
        assert usage.ru_idrss >= 0
        assert usage.ru_isrss >= 0
        assert usage.ru_minflt >= 0
        assert usage.ru_majflt >= 0
        assert usage.ru_nswap >= 0
        assert usage.ru_inblock >= 0
        assert usage.ru_oublock >= 0
        assert usage.ru_msgsnd >= 0
        assert usage.ru_msgrcv >= 0
        assert usage.ru_nsignals >= 0
        assert usage.ru_nvcsw >= 0
        assert usage.ru_nivcsw >= 0

    @pytest.mark.skipif(
        not is_windows(),
        reason="Avoiding Windows resource usage collection on Unix-like systems.",
    )
    def test_get_resource_usage_windows(self):
        """Test that get_resource_usage calls _get_windows_resource_usage on Windows systems"""
        usage = get_resource_usage()

        assert usage.ru_utime >= 0.0
        assert usage.ru_stime >= 0.0
        assert usage.ru_maxrss >= 0
        assert usage.ru_ixrss == 0
        assert usage.ru_idrss == 0
        assert usage.ru_isrss == 0
        assert usage.ru_minflt == 0
        assert usage.ru_majflt == 0
        assert usage.ru_nswap == 0
        assert usage.ru_inblock >= 0
        assert usage.ru_oublock >= 0
        assert usage.ru_msgsnd == 0
        assert usage.ru_msgrcv == 0
        assert usage.ru_nsignals == 0
        assert usage.ru_nvcsw >= 0
        assert usage.ru_nivcsw >= 0

    @pytest.mark.skipif(
        is_windows(),
        reason="Avoiding Unix resource usage collection on Windows. systems",
    )
    def test_get_unix_resource_usage(self):
        """Test _get_unix_resource_usage function"""
        usage = _get_unix_resource_usage()

        assert usage.ru_utime >= 0.0
        assert usage.ru_stime >= 0.0
        assert usage.ru_maxrss >= 0
        assert usage.ru_ixrss >= 0
        assert usage.ru_idrss >= 0
        assert usage.ru_isrss >= 0
        assert usage.ru_minflt >= 0
        assert usage.ru_majflt >= 0
        assert usage.ru_nswap >= 0
        assert usage.ru_inblock >= 0
        assert usage.ru_oublock >= 0
        assert usage.ru_msgsnd >= 0
        assert usage.ru_msgrcv >= 0
        assert usage.ru_nsignals >= 0
        assert usage.ru_nvcsw >= 0
        assert usage.ru_nivcsw >= 0

    @pytest.mark.skipif(
        not is_windows(),
        reason="Avoiding Windows resource usage collection on Unix-like systems.",
    )
    def test_get_windows_resource_usage_with_psutil(self):
        """Test _get_windows_resource_usage function with psutil available"""
        usage = _get_windows_resource_usage()

        assert usage.ru_utime >= 0.0
        assert usage.ru_stime >= 0.0
        assert usage.ru_maxrss >= 0
        assert usage.ru_ixrss == 0
        assert usage.ru_idrss == 0
        assert usage.ru_isrss == 0
        assert usage.ru_minflt == 0
        assert usage.ru_majflt == 0
        assert usage.ru_nswap == 0
        assert usage.ru_inblock >= 0
        assert usage.ru_oublock >= 0
        assert usage.ru_msgsnd == 0
        assert usage.ru_msgrcv == 0
        assert usage.ru_nsignals == 0
        assert usage.ru_nvcsw >= 0
        assert usage.ru_nivcsw >= 0

    @pytest.mark.skipif(
        not is_windows(),
        reason="Avoiding Windows resource usage collection on Unix-like systems.",
    )
    def test_get_windows_resource_usage_without_psutil(self, mocker):
        """Test _get_windows_resource_usage function when psutil is not available"""

        mocker.patch("psutil.Process", side_effect=ImportError)
        result = _get_windows_resource_usage()

        # Should return default ResourceUsage with all zeros
        assert result.ru_utime == 0.0
        assert result.ru_stime == 0.0
        assert result.ru_maxrss == 0
        assert result.ru_ixrss == 0
        assert result.ru_idrss == 0
        assert result.ru_isrss == 0
        assert result.ru_minflt == 0
        assert result.ru_majflt == 0
        assert result.ru_nswap == 0
        assert result.ru_inblock == 0
        assert result.ru_oublock == 0
        assert result.ru_msgsnd == 0
        assert result.ru_msgrcv == 0
        assert result.ru_nsignals == 0
        assert result.ru_nvcsw == 0
        assert result.ru_nivcsw == 0


# Made with Bob



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright IBM Corp. 2025

import logging

from wsgiref.simple_server import make_server
from spyne import Application, rpc, ServiceBase, Iterable, UnsignedInteger, \
    String, Unicode

from spyne.protocol.json import JsonDocument
from spyne.protocol.http import HttpRpc
from spyne.server.wsgi import WsgiApplication

from spyne.error import ResourceNotFoundError

from tests.helpers import testenv

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)

testenv["spyne_port"] = 10818
testenv["spyne_server"] = ("http://127.0.0.1:" + str(testenv["spyne_port"]))

class HelloWorldService(ServiceBase):
    @rpc(String, UnsignedInteger, _returns=Iterable(String))
    def say_hello(ctx, name, times):
        """
        :param name: The name to say hello to
        :param times: The number of times to say hello

        :returns: An array of 'Hello, <name>' strings, repeated <times> times.
        """

        for i in range(times):
            yield 'Hello, %s' % name

    @rpc(_returns=Unicode)
    def hello(ctx):
        return "<center><h1>🐍 Hello Stan! 🦄</h1></center>"
    
    @rpc(_returns=Unicode)
    def response_headers(ctx):
        ctx.transport.add_header("X-Capture-This", "this")
        ctx.transport.add_header("X-Capture-That", "that")
        return "Stan wuz here with headers!"
    
    @rpc(UnsignedInteger)
    def custom_404(ctx, user_id):
        raise ResourceNotFoundError(user_id)
    
    @rpc()
    def exception(ctx):
        raise Exception('fake error')


application = Application([HelloWorldService], 'instana.spyne.service.helloworld',
    in_protocol=HttpRpc(validator='soft'),
    out_protocol=JsonDocument(ignore_wrappers=True),
)
wsgi_app = WsgiApplication(application)
spyne_server = make_server('127.0.0.1', testenv["spyne_port"], wsgi_app)

if __name__ == '__main__':
    spyne_server.request_queue_size = 20
    spyne_server.serve_forever()



# (c) Copyright IBM Corp. 2025


try:
    import contextvars
    from typing import Any, Callable, Dict, List, Optional, Tuple

    import confluent_kafka  # noqa: F401
    import wrapt
    from confluent_kafka import Consumer, Producer
    from opentelemetry import context, trace
    from opentelemetry.trace import SpanKind

    from instana.log import logger
    from instana.propagators.format import Format
    from instana.singletons import get_tracer
    from instana.util.traceutils import (
        get_tracer_tuple,
        tracing_is_off,
    )
    from instana.span.span import InstanaSpan

    consumer_token = None
    consumer_span = contextvars.ContextVar("confluent_kafka_consumer_span")

    # As confluent_kafka is a wrapper around the C-developed librdkafka
    # (provided automatically via binary wheels), we have to create new classes
    # inheriting from the confluent_kafka package with the methods to be
    # monkey-patched.
    class InstanaConfluentKafkaProducer(Producer):
        """
        Wrapper class for confluent_kafka.Producer, which is an Asynchronous Kafka Producer.
        """

        def produce(
            self,
            topic: str,
            *args: object,
            **kwargs: Dict[str, Any],
        ) -> None:
            return super().produce(topic, *args, **kwargs)

    class InstanaConfluentKafkaConsumer(Consumer):
        """
        Wrapper class for confluent_kafka.Consumer, which is a high-level Apache Kafka consumer.
        """

        def consume(
            self, *args: object, **kwargs: Dict[str, Any]
        ) -> List[confluent_kafka.Message]:
            return super().consume(*args, **kwargs)

        def poll(
            self, timeout: Optional[float] = -1
        ) -> Optional[confluent_kafka.Message]:
            return super().poll(timeout)

        def close(self) -> None:
            return super().close()

    def trace_kafka_produce(
        wrapped: Callable[..., InstanaConfluentKafkaProducer.produce],
        instance: InstanaConfluentKafkaProducer,
        args: Tuple[int, str, Tuple[Any, ...]],
        kwargs: Dict[str, Any],
    ) -> None:
        if tracing_is_off():
            return wrapped(*args, **kwargs)

        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None
        is_suppressed = tracer.exporter._HostAgent__is_endpoint_ignored(
            "kafka",
            "produce",
            args[0],
        )

        with tracer.start_as_current_span(
            "kafka-producer", span_context=parent_context, kind=SpanKind.PRODUCER
        ) as span:
            span.set_attribute("kafka.service", args[0])
            span.set_attribute("kafka.access", "produce")

            # context propagation
            #
            # As stated in the official documentation at
            # https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#pythonclient-producer,
            # headers can be either a list of (key, value) pairs or a
            # dictionary. To maintain compatibility with the headers for the
            # Kafka Python library, we will use a list of tuples.
            headers = args[6] if len(args) > 6 else kwargs.get("headers", [])
            suppression_header = {"x_instana_l_s": "0" if is_suppressed else "1"}
            headers.append(suppression_header)

            tracer.inject(
                span.context,
                Format.KAFKA_HEADERS,
                headers,
                disable_w3c_trace_context=True,
            )

            headers.remove(suppression_header)

            if tracer.exporter.options.kafka_trace_correlation:
                kwargs["headers"] = headers
            try:
                res = wrapped(*args, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
            else:
                return res

    def create_span(
        span_type: str,
        topic: Optional[str] = "",
        headers: Optional[List[Tuple[str, bytes]]] = [],
        exception: Optional[str] = None,
    ) -> None:
        try:
            span = consumer_span.get(None)
            if span is not None:
                close_consumer_span(span)

            tracer, parent_span, _ = get_tracer_tuple()

            if not tracer:
                tracer = get_tracer()
            is_suppressed = False

            if topic:
                is_suppressed = tracer.exporter._HostAgent__is_endpoint_ignored(
                    "kafka",
                    span_type,
                    topic,
                )

            if not is_suppressed and headers:
                for header_name, header_value in headers:
                    if header_name == "x_instana_l_s" and header_value == b"0":
                        is_suppressed = True
                        break

            if is_suppressed:
                return

            parent_context = (
                parent_span.get_span_context()
                if parent_span
                else (
                    tracer.extract(
                        Format.KAFKA_HEADERS,
                        headers,
                        disable_w3c_trace_context=True,
                    )
                    if tracer.exporter.options.kafka_trace_correlation
                    else None
                )
            )
            span = tracer.start_span(
                "kafka-consumer", span_context=parent_context, kind=SpanKind.CONSUMER
            )
            if topic:
                span.set_attribute("kafka.service", topic)
            span.set_attribute("kafka.access", span_type)
            if exception:
                span.record_exception(exception)
                span.end()

            save_consumer_span_into_context(span)
        except Exception as e:
            logger.debug(
                f"Error while creating kafka-consumer span: {e}"
            )  # pragma: no cover

    def save_consumer_span_into_context(span: "InstanaSpan") -> None:
        global consumer_token
        ctx = trace.set_span_in_context(span)
        consumer_token = context.attach(ctx)
        consumer_span.set(span)

    def close_consumer_span(span: "InstanaSpan") -> None:
        global consumer_token
        if span.is_recording():
            span.end()
            consumer_span.set(None)
        if consumer_token is not None:
            context.detach(consumer_token)
            consumer_token = None

    def clear_context() -> None:
        global consumer_token
        context.attach(trace.set_span_in_context(None))
        consumer_token = None
        consumer_span.set(None)

    def trace_kafka_consume(
        wrapped: Callable[..., InstanaConfluentKafkaConsumer.consume],
        instance: InstanaConfluentKafkaConsumer,
        args: Tuple[int, str, Tuple[Any, ...]],
        kwargs: Dict[str, Any],
    ) -> List[confluent_kafka.Message]:
        res = None
        exception = None

        try:
            res = wrapped(*args, **kwargs)
            for message in res:
                create_span("consume", message.topic(), message.headers())
            return res
        except Exception as exc:
            exception = exc
            create_span("consume", exception=exception)

    def trace_kafka_close(
        wrapped: Callable[..., InstanaConfluentKafkaConsumer.close],
        instance: InstanaConfluentKafkaConsumer,
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
    ) -> None:
        try:
            # Close any existing consumer span before closing the consumer
            span = consumer_span.get(None)
            if span is not None:
                close_consumer_span(span)

            # Execute the actual close operation
            res = wrapped(*args, **kwargs)

            logger.debug("Kafka consumer closed and spans cleaned up")
            return res

        except Exception:
            # Still try to clean up the span even if close fails
            span = consumer_span.get(None)
            if span is not None:
                close_consumer_span(span)

    def trace_kafka_poll(
        wrapped: Callable[..., InstanaConfluentKafkaConsumer.poll],
        instance: InstanaConfluentKafkaConsumer,
        args: Tuple[int, str, Tuple[Any, ...]],
        kwargs: Dict[str, Any],
    ) -> Optional[confluent_kafka.Message]:
        res = None
        exception = None

        try:
            res = wrapped(*args, **kwargs)
            create_span("poll", res.topic(), res.headers())
            return res
        except Exception as exc:
            exception = exc
            create_span(
                "poll",
                next(iter(instance.list_topics().topics)),
                exception=exception,
            )

    # Apply the monkey patch
    confluent_kafka.Producer = InstanaConfluentKafkaProducer
    confluent_kafka.Consumer = InstanaConfluentKafkaConsumer

    wrapt.wrap_function_wrapper(
        InstanaConfluentKafkaProducer, "produce", trace_kafka_produce
    )
    wrapt.wrap_function_wrapper(
        InstanaConfluentKafkaConsumer, "consume", trace_kafka_consume
    )
    wrapt.wrap_function_wrapper(InstanaConfluentKafkaConsumer, "poll", trace_kafka_poll)
    wrapt.wrap_function_wrapper(
        InstanaConfluentKafkaConsumer, "close", trace_kafka_close
    )

    logger.debug("Instrumenting Kafka (confluent_kafka)")
except ImportError:
    pass
